from typing import Dict, List, Union

import numpy as np
import pandas as pd
import scipy
from numba import jit

from eyefeatures.preprocessing._utils import _get_distance, _get_MEC
from eyefeatures.preprocessing.base import BaseFixationPreprocessor


# ======== FIXATION PREPROCESSORS ========
class IVT(BaseFixationPreprocessor):
    """
    Velocity Threshold Identification.
    Complexity: O(n) for single group.
    """

    def __init__(
        self,
        x: str,
        y: str,
        t: str,
        threshold: float,
        distance: str = "euc",
        pk: List[str] = None,
        eps: float = 1e-10,
    ):
        super().__init__(x=x, y=y, t=t, pk=pk)
        self.threshold = threshold
        self.distance = distance
        self.eps = eps

        self.available_distances = ("euc", "manhattan", "chebyshev")  # TODO add more

    def _check_params(self):
        m = "IVT"
        assert self.x is not None, self._err_no_field(m, "x")
        assert self.y is not None, self._err_no_field(m, "y")
        assert self.t is not None, self._err_no_field(m, "t")
        assert self.distance in self.available_distances, (
            f"'distance' must be one of ({', '.join(self.available_distances)}),"
            f"got {self.distance}."
        )

    # @jit(forceobj=True, looplift=True)
    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        points = X[[self.x, self.y]].values
        dist = self._get_distances(points, self.distance)

        x = X[self.x].values
        y = X[self.y].values
        t = X[self.t].values

        dt = np.diff(t)
        vel = dist / (dt + self.eps)

        is_fixation = (vel <= self.threshold).astype(np.int32)
        fixations = self._squash_fixations(is_fixation)

        fixations_df = pd.DataFrame(
            data={
                "fixation_id": fixations,
                self.x: x[:-1],
                self.y: y[:-1],
                "start_time": t[:-1],
                "end_time": t[:-1],
                "distance_min": dist,
                "distance_max": dist,
            }
        )

        fixations_df = fixations_df[fixations_df["fixation_id"] != 0]
        diameters = []
        centers = []

        for i in set(fixations):
            if i != 0:
                points = fixations_df.loc[
                    fixations_df["fixation_id"] == i,
                    fixations_df.columns.isin([self.x, self.y]),
                ].values
                x, y, radius = _get_MEC(np.unique(points, axis=0))
                diameters.append(radius * 2)
                centers.append(np.array([x, y]))

        fixations_df = fixations_df.groupby(by=["fixation_id"]).agg(
            {
                self.x: "mean",
                self.y: "mean",
                "start_time": "min",
                "end_time": "max",
                "distance_min": "min",  # between consecutive gazes
                "distance_max": "max",
            }
        )

        fixations_df["diameters"] = diameters
        fixations_df["centers"] = centers

        # default features
        feats = (
            "duration",
            "saccade_duration",
            "saccade_length",
            "saccade_angle",
            "saccade2_angle",
        )
        fixations_df = self._compute_feats(fixations_df, feats)

        return fixations_df


class IDT(BaseFixationPreprocessor):
    """
    Dispersion Threshold Identification.
    Complexity: O(n * W)  for single group, where W is size of maximum window.
    Worst case is O(n^2) for W = n.
    """

    def __init__(
        self,
        x: str,
        y: str,
        t: str,
        min_duration: float,
        max_dispersion: float,
        distance: str = "euc",  # norm in R^2 for distance calculation
        pk: List[str] = None,
        eps: float = 1e-20,
    ):
        super().__init__(x=x, y=y, t=t, pk=pk)
        self.min_duration = min_duration
        self.max_dispersion = max_dispersion
        self.distance = distance
        self.eps = eps

        self.available_distances = ("euc", "manhattan", "chebyshev")  # TODO add more

    def _check_params(self):
        m = "IDT"
        assert self.x is not None, self._err_no_field(m, "x")
        assert self.y is not None, self._err_no_field(m, "y")
        assert self.t is not None, self._err_no_field(m, "t")
        assert self.min_duration is not None, self._err_no_field(m, "min_duration")
        assert self.min_duration > 0, f"'min_duration' must be non-negative."
        assert self.max_dispersion is not None, self._err_no_field(m, "min_duration")
        assert self.max_dispersion > 0, f"'max_dispersion' must be non-negative."

        assert self.distance in self.available_distances, (
            f"'distance' must be one of ({', '.join(self.available_distances)}),"
            f"got {self.distance}."
        )

    # @jit(forceobj=True, looplift=True)
    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        points = X[[self.x, self.y]].values
        dist = self._get_distances(points, self.distance)

        x = X[self.x].values
        y = X[self.y].values
        t = X[self.t].values

        dt = np.diff(t)
        vel = dist / (dt + self.eps)

        n = len(vel)
        is_fixation = np.zeros(n)

        cur = 0
        last = 0  # [cur, last]
        disp = np.full(n, -np.inf)

        while cur < n:
            # initiate time for window
            end_time = t[cur] + self.min_duration

            # invariant:
            # 1. current window's dispersion <= max_dispersion
            # 2. current window's duration < min_duration
            # so, algorithm classifies first correct window as fixation
            # TODO: implement approaches: widest possible window,
            #  some other characteristic

            # seek for first point to exceed end_time
            skip_gazes = False
            window_disp = -np.inf
            while last < n - 1 and t[last] <= end_time:
                next_idx = last + 1
                next_point = np.array(x[next_idx], y[next_idx])

                # update distances in window
                next_point_disp = -np.inf
                j = (
                    -1
                )  # index of gaze with which 'last'-th gaze achieved max dispersion
                for i in range(cur, next_idx):  # [cur, next_idx) == [cur, last]
                    d = _get_distance(
                        np.array(x[i], y[i]), next_point, distance=self.distance
                    )

                    disp[i] = max(disp[i], d)
                    if next_point_disp < d:
                        next_point_disp = d
                        j = i

                # check for dispersion
                if next_point_disp > self.max_dispersion:
                    # gazes [cur, j] cannot be part of fixation -> skip them
                    disp[j + 1 : next_idx] = -np.inf
                    cur = j + 1
                    last = j + 1
                    skip_gazes = True
                    break

                window_disp = max(window_disp, next_point_disp)
                disp[next_idx] = next_point_disp
                last = next_idx  # last = last + 1

            if skip_gazes:
                continue

            # if (last + 1 == len(df)) and there is not enough points
            if t[last] <= end_time:
                break

            # here we have correct window with duration >= min_duration
            # and dispersion <= max_dispersion.
            # Window could be extended further.
            is_fixation[cur : last + 1] = 1  # [cur, last] is single fixation
            disp[cur : last + 1] = window_disp
            cur = last + 1
            last = last + 1

        fixations = self._squash_fixations(is_fixation)

        fixations_df = pd.DataFrame(
            data={
                "fixation_id": fixations,
                self.x: x[:-1],
                self.y: y[:-1],
                "start_time": t[:-1],
                "end_time": t[:-1],
                "distance_min": dist,
                "distance_max": dist,
                "dispersion": disp,
            }
        )

        fixations_df = fixations_df[fixations_df["fixation_id"] != 0]
        diameters = []
        centers = []

        for i in set(fixations):
            if i != 0:
                points = fixations_df.loc[
                    fixations_df["fixation_id"] == i,
                    fixations_df.columns.isin([self.x, self.y]),
                ].values
                x, y, radius = _get_MEC(np.unique(points, axis=0))
                diameters.append(radius * 2)
                centers.append(np.array([x, y]))

        fixations_df = fixations_df.groupby(by=["fixation_id"]).agg(
            {
                self.x: "mean",
                self.y: "mean",
                "start_time": "min",
                "end_time": "max",
                "distance_min": "min",
                "distance_max": "max",
                "dispersion": "max",  # just for API, window has same values
            }
        )

        fixations_df["diameters"] = diameters
        fixations_df["centers"] = centers

        # default features
        feats = (
            "duration",
            "saccade_duration",
            "saccade_length",
            "saccade_angle",
            "saccade2_angle",
        )
        fixations_df = self._compute_feats(fixations_df, feats)

        return fixations_df


class IHMM(BaseFixationPreprocessor):
    """
    Hidden Markov Model Identification.
    Complexity: O(n^2) for single group.

    :param fix2sac: probability of transition from fixation to saccade.
    :param sac2fix: probability of transition from saccade to fixation.
    :param fix_distrib: distribution of fixations.
    :param sac_distrib: distribution of saccades.
    :param distrib_params: 'auto' for default params and dict {"fixation": params1, "saccade": params2}, where
           "params" are arguments for  appropriate `scipy.stats` function.
    """

    def __init__(
        self,
        x: str,
        y: str,
        t: str,
        fix2sac: float = 0.05,
        sac2fix: float = 0.05,
        fix_distrib: str = "norm",  # fixation distribution
        sac_distrib: str = "norm",
        distrib_params: Union[str, Dict[str, float]] = "auto",
        distance: str = "euc",
        pk: List[str] = None,
        eps: float = 1e-20,
    ):
        super().__init__(x=x, y=y, t=t, pk=pk)
        self.fix2sac = fix2sac
        self.sac2fix = sac2fix

        self.fix_distrib = fix_distrib
        self.sac_distrib = sac_distrib
        self.distrib_params = distrib_params
        self.distance = distance
        self.eps = eps

        self.available_distances = ("euc", "manhattan", "chebyshev")
        self.available_distributions = ("norm",)  # TODO: add more

    def _check_params(self):
        m = "IHMM"
        assert self.x is not None, self._err_no_field(m, "x")
        assert self.y is not None, self._err_no_field(m, "y")
        assert self.t is not None, self._err_no_field(m, "t")

        # TODO wrap this error message
        assert self.distance in self.available_distances, (
            f"'distance' must be one of ({', '.join(self.available_distances)}),"
            f"got '{self.distance}'."
        )
        assert self.fix_distrib in self.available_distributions, (
            f"'fix_distrib' must be one of ({', '.join(self.available_distributions)}),"
            f"got '{self.sac_distrib}'."
        )
        assert self.sac_distrib in self.available_distributions, (
            f"'sac_distrib' must be one of ({', '.join(self.available_distributions)}),"
            f"got '{self.sac_distrib}'."
        )
        assert isinstance(self.fix2sac, float) and (
            0.0 < self.fix2sac < 1.0
        ), f"'fix2sac' must be float between 0.0 and 1.0."
        assert isinstance(self.sac2fix, float) and (
            0.0 < self.sac2fix < 1.0
        ), f"'sac2fix' must be float between 0.0 and 1.0."

    def _get_distribution(self, ed, ep):
        if ed == "norm":
            return scipy.stats.norm(**ep)

        raise NotImplementedError(f"Distribution '{ed}' is not supported.")

    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        points = X[[self.x, self.y]].values
        dist = self._get_distances(points, self.distance)

        x = X[self.x].values
        y = X[self.y].values
        t = X[self.t].values

        dt = np.diff(t)
        vel = dist / (dt + self.eps)

        states = ("fixation", "saccade")
        start_probs = {"fixation": 0.5, "saccade": 0.5}

        transition_probs = {
            "fixation": {"fixation": 1 - self.fix2sac, "saccade": self.fix2sac},
            "saccade": {"fixation": self.sac2fix, "saccade": 1 - self.sac2fix},
        }

        # fixations have lower velocities than saccades
        if self.distrib_params == "auto":
            # TODO commented approach does not work
            # quantiles = np.quantile(vel, q=[0.15, 0.35, 0.55, 0.85])
            # fixation_mean = quantiles[0]                                   # 20th quantile as fixation mean estimate
            # fixation_std = vel[vel <= quantiles[1]].std()                  # [0, 40]'s std as fixation std estimate
            #
            # dp = {
            #     "fixation": {
            #         "loc": fixation_mean,
            #         "scale": fixation_std
            #     },
            #     "saccade": {
            #         "loc": vel.mean(),
            #         "scale": vel.std()
            #     }
            # }
            dp = {
                "fixation": {"loc": 0.002, "scale": 0.001},
                "saccade": {"loc": 0.03, "scale": 0.005},
            }
        else:
            dp = self.distrib_params

        fix_vel_distr = self._get_distribution(self.fix_distrib, dp["fixation"])
        sac_vel_distr = self._get_distribution(self.sac_distrib, dp["saccade"])

        emission_probs = {
            "fixation": lambda velocity: 1 - fix_vel_distr.cdf(velocity),
            "saccade": lambda velocity: sac_vel_distr.cdf(velocity),
        }

        _, best_path = self._viterbi(
            observations=vel,
            states=states,
            sp=start_probs,
            tp=transition_probs,
            ep=emission_probs,
        )

        is_fixation = (best_path == "fixation").astype(np.int32)

        fixations = self._squash_fixations(is_fixation)

        fixations_df = pd.DataFrame(
            data={
                "fixation_id": fixations,
                self.x: x[:-1],
                self.y: y[:-1],
                "start_time": t[:-1],
                "end_time": t[:-1],
                "distance_min": dist,
                "distance_max": dist,
            }
        )

        fixations_df = fixations_df[fixations_df["fixation_id"] != 0]
        diameters = []
        centers = []

        for i in set(fixations):
            if i != 0:
                points = fixations_df.loc[
                    fixations_df["fixation_id"] == i,
                    fixations_df.columns.isin([self.x, self.y]),
                ].values
                x, y, radius = _get_MEC(np.unique(points, axis=0))
                diameters.append(radius * 2)
                centers.append(np.array([x, y]))

        fixations_df = fixations_df.groupby(by=["fixation_id"]).agg(
            {
                self.x: "mean",
                self.y: "mean",
                "start_time": "min",
                "end_time": "max",
                "distance_min": "min",
                "distance_max": "max",
            }
        )

        fixations_df["diameters"] = diameters
        fixations_df["centers"] = centers

        # default features
        feats = (
            "duration",
            "saccade_duration",
            "saccade_length",
            "saccade_angle",
            "saccade2_angle",
        )
        fixations_df = self._compute_feats(fixations_df, feats)

        return fixations_df

    def _viterbi(self, observations, states, sp, tp, ep):
        """
        Computes hidden states vector Q s.t. probability
        of observing vector 'observations' given sp, tp, and ep is maximum.\n
        st ~ states\n
        sp ~ start_probs\n
        tp ~ transition_probs\n
        ep ~ emission_probs\n
        observations ~ velocities\n
        """
        prev_probs_nlog = None
        best_path = []
        if len(observations) == 0:
            return None

        for i, o_i in enumerate(observations):
            cur_probs_nlog = {}

            for j, q in enumerate(states):
                if i == 0:
                    q_prob_nlog_max = -np.log(max(sp[q], self.eps))
                else:
                    q_prob_nlog_max = max(
                        [prev_probs_nlog[q_] - np.log(tp[q_][q]) for q_ in states]
                    )

                cur_probs_nlog[q] = -np.log(max(ep[q](o_i), self.eps)) + q_prob_nlog_max

            best_path.append(max(cur_probs_nlog.keys(), key=cur_probs_nlog.get))
            prev_probs_nlog = cur_probs_nlog

        best_path_prob = max(np.exp(-prev_probs_nlog[q]) for q in states)
        return best_path_prob, np.array(best_path)



from abc import ABC, abstractmethod
from typing import Any, List, Tuple, Union

import numpy as np
import pandas as pd
from numba import jit, prange
from numpy.typing import NDArray
from scipy.stats import gaussian_kde
from sklearn.base import BaseEstimator, TransformerMixin

from eyefeatures.preprocessing._utils import _get_distance
from eyefeatures.utils import _get_angle, _get_angle3, _split_dataframe


class BasePreprocessor(BaseEstimator, TransformerMixin):
    def __init__(self, pk: List[str] = None):
        self.pk = pk

    @abstractmethod
    def _check_params(self):
        """
        Method checks that provided data is sufficient to conduct preprocessing.
        """
        ...

    @abstractmethod
    def _preprocess(self, X: pd.DataFrame):
        """
        Main method to preprocess fixations.
        """
        ...

    @staticmethod
    def _err_no_field(m, c):
        return f"Method '{m}' requires '{c}' for preprocessing."

    @jit(forceobj=True, looplift=True)
    def fit(self, X: pd.DataFrame, y=None):
        self._check_params()
        return self

    @jit(forceobj=True, looplift=True)
    def transform(self, X: pd.DataFrame) -> Union[pd.DataFrame, NDArray]:
        self._check_params()
        if self.pk is None:
            fixations = self._preprocess(X)
        else:
            fixations = None
            groups: List[str, pd.DataFrame] = _split_dataframe(X, self.pk, encode=False)
            for group_ids, group_X in groups:
                cur_fixations = self._preprocess(group_X)

                for i in prange(len(self.pk)):
                    cur_fixations.insert(loc=i, column=self.pk[i], value=group_ids[i])

                if fixations is None:
                    fixations = cur_fixations
                else:
                    fixations = pd.concat(
                        [fixations, cur_fixations], ignore_index=True, axis=0
                    )

        return fixations


class BaseFixationPreprocessor(BasePreprocessor, ABC):
    def __init__(self, x: str, y: str, t: str, pk: List[str] = None):
        super().__init__(pk=pk)
        self.x = x
        self.y = y
        self.t = t

    @staticmethod
    def _squash_fixations(is_fixation: NDArray) -> NDArray:
        """
        :param is_fixation: 0/1 array, whether a gaze is part of fixation.
        :returns: array of same size, ones are replaced with fixation_id.
        """
        n = len(is_fixation)
        fixation_id = 0
        prev_is_fixation = False
        for i in prange(n):
            if is_fixation[i] == 0:
                prev_is_fixation = False
                continue
            if not prev_is_fixation:
                fixation_id += 1
                prev_is_fixation = True

            is_fixation[i] = fixation_id

        return is_fixation

    @staticmethod
    def _get_distances(points: NDArray, distance):
        dist = np.zeros(len(points) - 1)
        for i in prange(len(points) - 1):
            dist[i] = _get_distance(points[i, :], points[i + 1, :], distance=distance)
        return dist

    def _compute_feats(
        self, fixations_df: pd.DataFrame, feats: Tuple[str, ...]
    ) -> pd.DataFrame:
        """
        Method computes list of required features.
        """
        n = len(fixations_df)

        # fixation duration
        if "duration" in feats:
            fixations_df["duration"] = fixations_df.end_time - fixations_df.start_time

        # saccade preceding the fixation
        if "saccade_duration" in feats:
            sd = fixations_df.start_time.values - fixations_df.end_time.shift(1).values
            sd[0] = 0  # no preceding saccade
            fixations_df["saccade_duration"] = sd

        # saccade preceding the fixation
        if "saccade_length" in feats:
            start_points = fixations_df[[self.x, self.y]].values
            end_points = fixations_df[[self.x, self.y]].shift(1).values
            sl = _get_distance(
                end_points,
                start_points,
                distance=self.distance,  # initialized by child class
            )
            sl[0] = 0
            fixations_df["saccade_length"] = sl

        # angle between x-axis and saccade preceding the fixation,
        # assuming that saccade moves from (0, 0) to current fixation
        if "saccade_angle" in feats:
            dx: pd.Series = fixations_df[self.x].diff().values
            dy: pd.Series = fixations_df[self.y].diff().values
            sa = np.zeros(shape=(n,))
            for i in prange(1, n):
                sa[i] = _get_angle(dx[i], dy[i], degrees=True)
            fixations_df["saccade_angle"] = sa

        # angle between preceding and succeeding saccades
        if "saccade2_angle" in feats:
            xx, yy = fixations_df[self.x].values, fixations_df[self.y].values
            sa2 = np.zeros(shape=(n,))
            for i in prange(1, n - 1):
                sa2[i] = _get_angle3(
                    x0=xx[i],
                    y0=yy[i],
                    x1=xx[i - 1],
                    y1=yy[i - 1],
                    x2=xx[i + 1],
                    y2=yy[i + 1],
                )
            fixations_df["saccade2_angle"] = sa2

        return fixations_df


class BaseAOIPreprocessor(BasePreprocessor, ABC):
    def __init__(self, x: str, y: str, t: str, aoi: str = None, pk: List[str] = None):
        super().__init__(pk=pk)
        self.x = x
        self.y = y
        self.t = t
        self.aoi = aoi

    def _get_fixation_density(
        self, data: pd.DataFrame
    ) -> tuple[np.ndarray[Any, np.dtype], Any, Any]:
        """
        Finds the fixation density of a given dataframe.

        :param data: DataFrame with fixations.
        :return: density for each point in [x_min, x_max] x [y_min, y_max] area
        """
        df = data[[self.x, self.y]]
        assert df.shape[0] > 2, "Not enough points"
        kde = gaussian_kde(df.values.T)
        X, Y = np.mgrid[
            df[self.x].min() : df[self.x].max() : 100j,
            df[self.y].min() : df[self.y].max() : 100j,
        ]  # is 100 enough?
        positions = np.vstack([X.ravel(), Y.ravel()])
        return np.reshape(kde(positions), X.shape), X, Y

    @staticmethod
    @jit(forceobj=True, looplift=True)
    def _find_local_max_coordinates(loc_max_matrix: np.ndarray) -> np.ndarray:
        """
        Finds the local max coordinates of a fixation density matrix.

        :param loc_max_matrix: matrix with maxima.
        """
        for i in prange(loc_max_matrix.shape[0]):  # TODO vectorize with numpy?
            for j in prange(loc_max_matrix.shape[1]):
                if i == 0 and j != 0:
                    if loc_max_matrix[i][j - 1] == loc_max_matrix[i][j]:
                        loc_max_matrix[i][j - 1] = 0
                elif j == 0 and i != 0:
                    if loc_max_matrix[i - 1][j] == loc_max_matrix[i][j]:
                        loc_max_matrix[i - 1][j] = 0
                elif i != 0 and j != 0:
                    if loc_max_matrix[i - 1][j] == loc_max_matrix[i][j]:
                        loc_max_matrix[i - 1][j] = 0
                    if loc_max_matrix[i - 1][j - 1] == loc_max_matrix[i][j]:
                        loc_max_matrix[i - 1][j - 1] = 0
                    if loc_max_matrix[i][j - 1] == loc_max_matrix[i][j]:
                        loc_max_matrix[i][j - 1] = 0
        return np.transpose(np.nonzero(loc_max_matrix))

    def _scale_coordinates(self, X: pd.DataFrame) -> pd.DataFrame:
        X[self.x] -= X[self.x].mean()
        X[self.y] -= X[self.y].mean()
        return X


class BaseSmoothingPreprocessor(BasePreprocessor, ABC):
    def __init__(self, x: str, y: str, t: str, pk: List[str] = None):
        super().__init__(pk=pk)
        self.x = x
        self.y = y
        self.t = t



import math
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from numba import jit
from scipy.ndimage import maximum_filter, prewitt, sobel
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin

from eyefeatures.features.measures import ShannonEntropy
from eyefeatures.preprocessing.base import BaseAOIPreprocessor
from eyefeatures.utils import _split_dataframe

# ======== AOI PREPROCESSORS ========


class ShapeBased(BaseAOIPreprocessor):
    """
    Defines AOI using the specified shapes

    :param x: x coordinate of fixation.
    :param y: y coordinate of fixation.
    :param aoi_name: name of AOI column.
    :param pk: list of column names used to split pd.DataFrame.
    :param shapes: list of shapes. It should be a list of tuple lists. Parameters for shape:\n
                    \n
                    0: 'r', 'c', 'e': rectangle, circle, ellipse\n
                    For the rectangle:\n
                    1: coordinates of the lower left corner of the rectangle.\n
                    2: coordinates of the upper right corner of the rectangle.\n
                    For the circle:\n
                    1: coordinates of the center of the circle.\n
                    2: radius of the circle.\n
                    For the ellipse:\n
                    ((x - x')*cos(alpha) + (y - y')*sin(alpha))**2 / a**2 + (-(x - x')*sin(alpha) + (y - y')*cos(alpha))**2 / b**2 = c\n
                    1: coordinates of the center of the ellipse (x', y').\n
                    2: "a" in the ellipse equation\n
                    3: "b" in the ellipse equation\n
                    4: "c" in the ellipse equation\n
                    5: angle of inclination of th ellipse in radians (alpha)\n

    """

    def __init__(
        self,
        x: str = None,
        y: str = None,
        shapes: List = None,
        aoi_name: str = "AOI",
        pk: List[str] = None,
    ):
        super().__init__(x=x, y=y, t=None, aoi=aoi_name, pk=None)
        self.shapes = shapes
        self.instance = pk

    def _check_params(self):
        m = "ShapeBased"
        assert self.x is not None, self._err_no_field(m, "x")
        assert self.y is not None, self._err_no_field(m, "y")
        assert self.shapes is not None, self._err_no_field(m, "shapes")

    def _is_inside_of_fig(self, X: pd.DataFrame, shape_id):
        ind = 0
        for shape in self.shapes[shape_id]:
            if shape[0] == "r":  # Rectangle
                X.loc[
                    (X[self.x] <= shape[2][0])
                    & (X[self.x] >= shape[1][0])
                    & (X[self.y] <= shape[2][1])
                    & (X[self.y] >= shape[1][1]),
                    self.aoi,
                ] = f"aoi_{ind}"
            elif shape[0] == "c":  # Circle
                X["length"] = X.apply(
                    lambda z: np.linalg.norm(
                        np.array((z[self.x], z[self.y])) - np.array(shape[1])
                    ),
                    axis=1,
                )
                X.loc[X["length"] <= shape[2], self.aoi] = f"aoi_{ind}"
            elif shape[0] == "e":  # Ellipse
                X.loc[
                    (
                        (X[self.x] - shape[1][0]) * np.cos(shape[5])
                        + (X[self.y] - shape[1][1]) * np.sin(shape[5])
                    )
                    ** 2
                    / (shape[2] ** 2)
                    + (
                        -(X[self.x] - shape[1][0]) * np.sin(shape[5])
                        + (X[self.y] - shape[1][1]) * np.cos(shape[5])
                    )
                    / (shape[3] ** 2)
                    <= shape[4],
                    self.aoi,
                ] = f"aoi_{ind}"
            ind += 1
        return X[self.aoi]

    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        assert X.shape[0] != 0, "Error: there are no points"

        # if self.instance is not None:
        #     X.drop(columns=self.instance, inplace=True)
        # X[self.aoi] = None
        to_concat = []
        flag = False
        if len(self.shapes) != 1:
            flag = True
        shape_id = 0
        if self.instance is None:
            X[self.aoi] = self._is_inside_of_fig(X, shape_id)
            fixations = X
        else:
            instances: List[str, pd.DataFrame] = _split_dataframe(
                X, self.instance, encode=False
            )
            assert (not flag) or len(instances) == len(self.shapes), "Not enough shapes"
            for instance_ids, instance_X in instances:
                instance_X[self.aoi] = self._is_inside_of_fig(instance_X, shape_id)
                to_concat.append(instance_X)
                if flag:
                    shape_id += 1
            fixations = pd.concat(to_concat, axis=0)

        return fixations


class ThresholdBased(BaseAOIPreprocessor):
    """
    Defines the AOI for each fixation using density maximum and Kmeans. Finds local maximum, pre-threshold it, and uses
    it as a center of aoi

    :param x: x coordinate of fixation.
    :param y: y coordinate of fixation.
    :param window_size: size of search window.
    :param threshold: threshold density.
    :param pk: list of column names used to split pd.DataFrame.
    :param aoi_name: name of AOI column.
    :param algorithm_type: type of clustering algorithm to use.
    :param threshold_dist:
    """

    def __init__(
        self,
        x: str = None,
        y: str = None,
        window_size: int = None,
        threshold: float = None,
        pk: List[str] = None,
        aoi_name: str = None,
        algorithm_type: str = "kmeans",
        threshold_dist: float = None,
    ):
        super().__init__(x=x, y=y, t=None, aoi=aoi_name, pk=pk)
        self.window_size = window_size
        self.threshold = threshold
        self.algorithm_type = algorithm_type
        self.threshold_dist = threshold_dist

    def _check_params(self):
        m = "ThresholdBased"
        assert self.x is not None, self._err_no_field(m, "x")
        assert self.y is not None, self._err_no_field(m, "y")
        assert self.window_size is not None, self._err_no_field(m, "window_size")
        assert self.threshold is not None, self._err_no_field(m, "threshold")
        assert self.window_size > 0, "Error: window size must be greater than zero"
        assert (
            self.threshold >= 0
        ), "Error: threshold must be greater than zero or equal to zero"
        assert self.algorithm_type in [
            "kmeans",
            "basic",
        ], "Error: only 'kmeans' or 'basic' are supported"
        if self.algorithm_type == "basic":
            assert self.threshold_dist is not None, self._err_no_field(
                m, "threshold_dist"
            )

    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        assert X.shape[0] != 0, "Error: there are no points"

        if self.pk is not None:
            X.drop(columns=self.pk, inplace=True)

        aoi_list = []
        density, X_grid, Y_grid = self._get_fixation_density(X)

        # For each sliding window (window_size x window_size) finds maximum density
        mx = maximum_filter(density, size=(self.window_size, self.window_size))
        # Filter local maxima (0 if maximus less than threshold)
        loc_max_matrix = np.where((mx == density) & (mx >= self.threshold), density, 0)
        loc_max_coord = super()._find_local_max_coordinates(loc_max_matrix)

        assert (
            loc_max_coord.shape[0] != 0
        ), "Error: Can't find the maximum with such parameters"

        aoi_counts: Dict[str, int] = dict()  # Dict[aoi name] = count of points in aoi
        aoi_points: Dict[str, List[Tuple[float, float]]] = (
            dict()
        )  # Dict[aoi name] = list of points

        axis_x = X_grid.T[0]
        axis_y = Y_grid[0]
        centers: Dict[str, Tuple[float, float]] = dict()  # Dict with the centers of aoi
        for i in range(loc_max_coord.shape[0]):  # Initial centers for each AOI
            centers[f"aoi_{i}"] = (
                X_grid[loc_max_coord[i][0]][0],
                Y_grid[loc_max_coord[i][1]][0],
            )
            aoi_points[f"aoi_{i}"] = [centers[f"aoi_{i}"]]

        for index, row in X.iterrows():
            min_dist = np.inf
            min_dist_aoi = None
            x_coord = min(np.searchsorted(axis_x, row[self.x]), 99)
            y_coord = min(np.searchsorted(axis_y, row[self.y]), 99)
            # Find the nearst aoi
            for key in centers.keys():
                if self.algorithm_type == "kmeans":
                    dist = np.linalg.norm(
                        np.array([row[self.x], row[self.y]]) - np.array(centers[key])
                    )
                    if dist < min_dist:
                        min_dist = dist
                        min_dist_aoi = key
                if self.algorithm_type == "basic":
                    length = np.inf
                    for point in aoi_points[
                        key
                    ]:  # Find minimal distance between fixation and each fixation in particular aoi
                        dist = np.linalg.norm(
                            np.array([row[self.x], row[self.y]]) - np.array(point)
                        )
                        length = min(length, dist)
                        if dist >= self.threshold_dist:
                            length = np.inf
                            break
                    if min_dist > length:
                        min_dist = length
                        min_dist_aoi = key

            # Add fixation to the aoi (basic algorithm)
            if (
                self.algorithm_type == "basic"
                and min_dist_aoi is not None
                and density[x_coord][y_coord] > self.threshold
            ):
                aoi_points[min_dist_aoi].append((row[self.x], row[self.y]))
                aoi_counts[min_dist_aoi] = len(aoi_points[min_dist_aoi]) - 1

            # Add fixation to the aoi (kmeans algorithm) and compute the new center
            if self.algorithm_type == "kmeans" and min_dist_aoi is not None:
                if (
                    min_dist_aoi in aoi_counts.keys()
                    and density[x_coord][y_coord] > self.threshold
                ):
                    aoi_counts[min_dist_aoi] += 1

                    new_center_x = (
                        centers[min_dist_aoi][0]
                        * (aoi_counts[min_dist_aoi] - 1)
                        / aoi_counts[min_dist_aoi]
                        + row[self.x] / aoi_counts[min_dist_aoi]
                    )
                    new_center_y = (
                        centers[min_dist_aoi][1]
                        * (aoi_counts[min_dist_aoi] - 1)
                        / aoi_counts[min_dist_aoi]
                        + row[self.y] / aoi_counts[min_dist_aoi]
                    )
                    centers[min_dist_aoi] = (new_center_x, new_center_y)
                elif (
                    min_dist_aoi not in aoi_counts.keys()
                    and density[x_coord][y_coord] > self.threshold
                ):
                    aoi_counts[min_dist_aoi] = 1
                    new_center_x = (centers[min_dist_aoi][0] + row[self.x]) / 2
                    new_center_y = (centers[min_dist_aoi][1] + row[self.y]) / 2
                    centers[min_dist_aoi] = (new_center_x, new_center_y)
                elif density[x_coord][y_coord] <= self.threshold:
                    min_dist_aoi = None

            aoi_list.append(min_dist_aoi)

        X[self.aoi] = aoi_list
        return X


class GradientBased(BaseAOIPreprocessor):
    """
    Defines the AOI for each fixation using a gradient-based algorithm. Finds the local maximum, pre-threshold it, and
    uses it as a center of aoi. After that, uses the Sobel operator to compute the gradient magnitude for each point.
    Next, defines the queue of areas of interest. Algorithm of aoi defining:\n
    * Gets the point from the queue. It is a center\n
    * Looks at the points near the center\n
    * Tries to find the point with defined aoi and maximum gradient magnitude\n
    * Adds center to this aoi\n
    * Repeats for all points in the matrix\n

    :param x: x coordinate of fixation.
    :param y: y coordinate of fixation.
    :param window_size: size of search window.
    :param threshold: threshold density.
    :param pk: list of column names used to split pd.DataFrame.
    :param aoi_name: name of AOI column.
    """

    def __init__(
        self,
        x: str = None,
        y: str = None,
        window_size: int = None,
        threshold: float = None,
        pk: List[str] = None,
        aoi_name: str = None,
    ):
        super().__init__(x=x, y=y, t=None, aoi=aoi_name, pk=pk)
        self.window_size = window_size
        self.threshold = threshold

    def _check_params(self):
        m = "GradientBased"
        assert self.x is not None, self._err_no_field(m, "x")
        assert self.y is not None, self._err_no_field(m, "y")
        assert self.window_size is not None, self._err_no_field(m, "window_size")
        assert self.threshold is not None, self._err_no_field(m, "threshold")
        assert self.window_size > 0, "Error: window size must be greater than zero"
        assert (
            self.threshold >= 0
        ), "Error: threshold must be greater than zero or equal to zero"

    # @jit(forceobj=True, looplift=True)
    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        assert X.shape[0] != 0, "Error: there are no points"
        if self.pk is not None:
            X.drop(columns=self.pk, inplace=True)

        aoi_list = []
        density, X_grid, Y_grid = super()._get_fixation_density(X)

        # For each sliding window (window_size x window_size) finds maximum density
        mx = maximum_filter(density, size=(self.window_size, self.window_size))
        # Filter local maxima (0 if maximus less than threshold)
        loc_max_matrix = np.where((mx == density) & (mx >= self.threshold), density, 0)
        loc_max_coord = super()._find_local_max_coordinates(loc_max_matrix)

        assert (
            loc_max_coord.shape[0] != 0
        ), "Error: Can't find the maximum with such parameters"

        axis_x = X_grid.T[0]
        axis_y = Y_grid[0]
        centers: Dict[str, Tuple[float, float]] = dict()  # Dict with the centers of aoi
        aoi_matrix = np.zeros((density.shape[0], density.shape[1]), dtype=int)
        # Compute the gradient magnitude
        horizontal_sobel = sobel(density, axis=0)
        vertical_sobel = sobel(density, axis=1)
        magnitude_sobel = np.sqrt(horizontal_sobel**2 + vertical_sobel**2)
        magnitude_sobel = np.pad(
            magnitude_sobel, 2, mode="constant", constant_values=-1
        )

        queue_of_centers: List[List[Tuple[int, int]]] = (
            []
        )  # List of points to add for each aoi
        for i in range(loc_max_coord.shape[0]):  # Initial centers for each AOI
            centers[f"aoi_{i}"] = (
                X_grid[loc_max_coord[i][0]][0],
                Y_grid[loc_max_coord[i][1]][0],
            )
            queue_of_centers.append([])
            for j in range(-1, 2):
                for k in range(-1, 2):
                    if (
                        0 <= loc_max_coord[i][0] + j < density.shape[0]
                        and 0 <= loc_max_coord[i][1] + k < density.shape[0]
                        and not (j == 0 and k == 0)
                        and aoi_matrix[loc_max_coord[i][0] + j][loc_max_coord[i][1] + k]
                        == 0
                    ):
                        queue_of_centers[-1].append(
                            (loc_max_coord[i][0] + j, loc_max_coord[i][1] + k)
                        )
            aoi_matrix[loc_max_coord[i][0]][loc_max_coord[i][1]] = i + 1
        ind = 0
        while any([len(x) > 0 for x in queue_of_centers]) > 0:
            # If the list of points to add for particular aoi is not empty, then try to add them to aoi, else this aoi
            # is built
            if len(queue_of_centers[ind]) > 0:
                x_coord, y_coord = queue_of_centers[ind].pop(
                    0
                )  # Get the point without aoi. It is a center of window
                if aoi_matrix[x_coord][y_coord] != 0:  # are all the fixation covered?
                    continue
                # Add 2 due to padding
                window_magnitude = magnitude_sobel[
                    x_coord + 1 : x_coord + 4, y_coord + 1 : y_coord + 4
                ]
                max_magnitude = -1
                aoi_to_add = None
                # Find the point in window with the biggest gradient magnitude and take its aoi
                for j in range(-1, 2):
                    for k in range(-1, 2):
                        if (
                            0 <= x_coord + j < density.shape[0]
                            and 0 <= y_coord + k < density.shape[0]
                            and not (j == 0 and k == 0)
                        ):
                            # If the point in the window (except for the center) has greater magnitude and aoi,
                            # then we take its aoi
                            if (
                                aoi_matrix[x_coord + j][y_coord + k] != 0
                                and max_magnitude <= window_magnitude[1 + j][1 + k]
                            ):
                                aoi_to_add = aoi_matrix[x_coord + j][y_coord + k]
                                max_magnitude = window_magnitude[1 + j][1 + k]
                            # If the point in the window (except for the center) doesn't have aoi, then add it to queue
                            elif (
                                aoi_matrix[x_coord + j][y_coord + k] == 0
                                and (x_coord + j, y_coord + k)
                                not in queue_of_centers[ind]
                            ):  # and magnitude[1+j][1+k] >= gradient_eps:
                                queue_of_centers[ind].append((x_coord + j, y_coord + k))
                aoi_matrix[x_coord][y_coord] = aoi_to_add
            ind = (ind + 1) % len(queue_of_centers)

        # Match fixations and aoi in aoi matrix
        for index, row in X.iterrows():
            x_coord = min(np.searchsorted(axis_x, row[self.x]), 99)
            y_coord = min(np.searchsorted(axis_y, row[self.y]), 99)
            if aoi_matrix[x_coord][y_coord] == 0:
                aoi_list.append(None)
            else:
                aoi_list.append(f"aoi_{aoi_matrix[x_coord][y_coord]}")

        X[self.aoi] = aoi_list
        return X


class OverlapCLustering(BaseAOIPreprocessor):
    """
    Defines the AOI for each fixation using the overlapping clustering algorithm.

    :param x: x coordinate of fixation.
    :param y: y coordinate of fixation.
    :param diameters: diameters of fixation.
    :param centers: centers of fixation.
    :param pk: list of column names used to split pd.DataFrame.
    :param aoi_name: name of AOI column.
    :param eps: additional length to sum of radius
    """

    def __init__(
        self,
        x: str = None,
        y: str = None,
        diameters: str = None,
        centers: str = None,
        pk: List[str] = None,
        aoi_name: str = None,
        eps: float = 0.0,
    ):
        super().__init__(x=x, y=y, t=None, aoi=aoi_name, pk=pk)
        self.diameters = diameters
        self.centers = centers
        self.eps = eps

    def _check_aoi(self):
        m = "OverlapCLustering"
        assert self.x is not None, self._err_no_field(m, "x")
        assert self.y is not None, self._err_no_field(m, "y")
        assert self.diameters is not None, self._err_no_field(m, "diameters")
        assert self.centers is not None, self._err_no_field(m, "centers")

    def _build_clusters(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        First step of the overlapping clustering algorithm. Builds the clusters. If the fixation locates inside another
        one, then these fixations are in one aoi.
        """
        X[self.aoi] = 0
        cluster_id = 1
        for index, row in X.iterrows():
            if X.loc[index, X.columns == self.aoi].values[0] == 0:
                X.loc[index, X.columns == self.aoi] = cluster_id
                center = row[self.centers]
                diameter = row[self.diameters]
                # Calculate the distance between centers
                X["diff_center"] = X[self.centers].apply(
                    lambda p: (p[0] - center[0]) ** 2 + (p[1] - center[1]) ** 2
                )
                X["diff_diam"] = abs(diameter - X[self.diameters]) / 2
                # Add all fixation, which are
                fixation_in_cluster = X[X["diff_center"] <= X["diff_diam"]].index
                X.loc[fixation_in_cluster, self.aoi] = cluster_id
                cluster_id += 1

        return X

    def _merge_clusters(self, X: pd.DataFrame) -> pd.DataFrame:
        """Second step of the overlapping clustering algorithm. Merges the clusters. Selects aoi with the most amount of
        fixation in itself. Creates the queue of fixation and starts the cycle of merging
        """
        used = []
        while len(X[~X[self.aoi].isin(used)]) > 0:
            # Find the largest cluster(aoi)
            max_cluster = (
                X[~X[self.aoi].isin(used)].groupby(self.aoi).count().idxmax().iloc[0]
            )
            points = X[X[self.aoi] == max_cluster].index.values.tolist()
            ind = 0
            end_ = len(points)
            used.append(max_cluster)
            while ind < end_:
                row = X.iloc[points[ind]]
                # Calculate the distance between centers
                X["length"] = X[self.centers].apply(
                    lambda p: np.linalg.norm(p - row[self.centers])
                )
                # Merge areas of interest that intersect the max cluster
                to_merge = X[
                    (
                        X["length"]
                        <= abs((X[self.diameters] + row[self.diameters]) / 2 + self.eps)
                    )
                    & (~X[self.aoi].isin(used))
                ][self.aoi].unique()
                add_fixations = X[
                    (X[self.aoi].isin(to_merge)) & (~X[self.aoi].isin(used))
                ].index.values
                X.loc[X[self.aoi].isin(to_merge), (X.columns == self.aoi)] = max_cluster
                # Add fixation from those areas of interest to queue
                points.extend(add_fixations)
                ind += 1
                end_ += len(add_fixations)

        return X

    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        X.drop(columns=self.pk, inplace=True)
        X.reset_index(drop=True, inplace=True)
        copy_X = X.copy()
        copy_X = self._build_clusters(copy_X)
        copy_X = self._merge_clusters(copy_X)
        X[self.aoi] = copy_X[self.aoi]
        return X


# ======== EXTRACTOR FOR AOI CLASSES ========
class AOIExtractor(BaseEstimator, TransformerMixin):
    """Extractor of areas of interest. Selects the partition into zones of interest with the lowest entropy.

    :param methods: list of aoi algorithms.
    :param x: x coordinate of fixation.
    :param y: y coordinate of fixation.
    :param window_size: size of search window.
    :param threshold: threshold density.
    :param pk: list of column names used to split pd.DataFrame for scaling.
    :param instance_columns: list of column names used to split pd.DataFrame into the similar instances for aoi extraction.
    :param aoi_name: name of AOI column.
    :param show_best: if true, then return the best method for each instance
    """

    def __init__(
        self,
        methods: List[BaseAOIPreprocessor],
        x: str,
        y: str,
        window_size: int = None,
        threshold: float = None,
        pk: List[str] = None,
        instance_columns: List[str] = None,
        aoi_name: str = None,
        show_best: bool = False,
    ):
        self.x = x
        self.y = y
        self.methods = methods
        self.window_size = window_size
        self.threshold = threshold
        self.pk = pk
        self.instance_columns = instance_columns
        self.aoi = aoi_name
        self.show_best = show_best

    # @jit(forceobj=True, looplift=True)
    def fit(self, X: pd.DataFrame, y=None):
        for method in self.methods:
            method.x = self.x
            method.y = self.y
            if self.window_size is not None:
                method.window_size = self.window_size
            if self.threshold is not None:
                method.threshold = self.threshold
            # method.pk = self.pk
            method.aoi = self.aoi
            if not isinstance(method, ClusterMixin):
                method.fit(X)
        return self

    # @jit(forceobj=True, looplift=True)
    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        if self.methods is None:
            return X

        data_df: pd.DataFrame = X[[self.x, self.y]]
        if self.pk is not None:
            data_df = pd.concat([data_df, X[self.pk]], axis=1)

        if self.instance_columns is not None:
            to_add = [x for x in self.instance_columns if x not in self.pk]
            data_df = pd.concat([data_df, X[to_add]], axis=1)
        else:
            self.instance_columns = self.pk

        fixations = None
        instances: List[str, pd.DataFrame] = _split_dataframe(
            X, self.instance_columns, encode=False
        )
        shapes_id = 0  # For ShapeBased
        # Entropy for selecting the best method
        entropy_transformer = ShannonEntropy(aoi=self.aoi, pk=self.instance_columns)
        # Extract areas of interest for each instance
        for instance_ids, instance_X in instances:
            min_entropy = np.inf
            fixations_with_aoi = None
            # Select best aoi extraction
            for method in self.methods:
                copy_x = None
                copy_y = None
                to_transform = None
                groups: List[str, pd.DataFrame] = _split_dataframe(
                    instance_X, self.pk, encode=False
                )
                # Scale each group of fixation
                for group_ids, group_X in groups:
                    if copy_x is None:
                        copy_x = group_X[self.x]
                        copy_y = group_X[self.y]
                    else:
                        copy_x = pd.concat(
                            [copy_x, group_X[self.x]], ignore_index=True, axis=0
                        )
                        copy_y = pd.concat(
                            [copy_y, group_X[self.y]], ignore_index=True, axis=0
                        )
                    group_X[self.x] -= group_X[self.x].mean()
                    group_X[self.y] -= group_X[self.y].mean()
                    if to_transform is None:
                        to_transform = group_X
                    else:
                        to_transform = pd.concat(
                            [to_transform, group_X], ignore_index=True, axis=0
                        )

                if isinstance(method, ClusterMixin):
                    cur_aoi = method.fit_predict(to_transform[[self.x, self.y]])
                    cur_fixations = pd.concat(
                        [to_transform, pd.Series(cur_aoi, name=self.aoi)], axis=1
                    )
                elif isinstance(method, ShapeBased):
                    save_shapes = method.shapes
                    assert (len(save_shapes) == 1) or (
                        len(save_shapes) != len(instances)
                    ), "Not enough shapes"
                    method.shapes = [
                        save_shapes[shapes_id],
                    ]
                    method.instance = None
                    cur_fixations = method.transform(to_transform)
                    method.shapes = save_shapes
                    if len(save_shapes) != 1:
                        shapes_id += 1
                else:
                    cur_fixations = method.transform(to_transform)

                all_areas = np.unique(
                    [el for el in cur_fixations[self.aoi].values if not el is None]
                )
                areas_names = [f"aoi_{i}" for i in range(len(all_areas))]
                map_areas = dict(zip(all_areas, areas_names))
                cur_fixations[self.aoi] = cur_fixations[self.aoi].map(map_areas)
                entropy = entropy_transformer.transform(cur_fixations)[
                    "entropy"
                ].values[0]
                if min_entropy > entropy:
                    min_entropy = entropy
                    fixations_with_aoi = cur_fixations
                    fixations_with_aoi[self.x] = copy_x
                    fixations_with_aoi[self.y] = copy_y
                    if self.show_best:
                        fixations_with_aoi["best_method"] = method.__class__.__name__
            if fixations is None:
                fixations = fixations_with_aoi
            else:
                fixations = pd.concat(
                    [fixations, fixations_with_aoi], ignore_index=True, axis=0
                )

        return fixations


class AOIMatcher(BaseEstimator, TransformerMixin):
    """
    Matches AOI in the dataset.

    :param x: x coordinate of fixation.
    :param y: y coordinate of fixation.
    :param pk: list of column names used to split pd.DataFrame for scaling.
    :param instance_columns: list of column names used to split pd.DataFrame
                             into the similar instances for aoi extraction.
    :param aoi: name of AOI column.
    :param n_aoi: count of aoi in the group\n
                 0: any number the areas of interest\n
                 all integer number that greater than 0: count of the areas of interest.
    """

    def __init__(
        self,
        x: str,
        y: str,
        pk: List[str] = None,
        instance_columns: List[str] = None,
        aoi: str = None,
        n_aoi: int = 0,
    ):
        self.x = x
        self.y = y
        self.pk = pk
        self.instance_columns = instance_columns
        self.aoi = aoi
        self.n_aoi = n_aoi

    def fit(self, X: pd.DataFrame, y=None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        data_df: pd.DataFrame = X.copy()

        fixations = None
        instances: List[str, pd.DataFrame] = _split_dataframe(
            data_df, self.instance_columns, encode=False
        )
        prev_pattern = dict()
        for instance_ids, instance_X in instances:
            copy_x = None
            copy_y = None
            groups: List[str, pd.DataFrame] = _split_dataframe(
                instance_X, self.pk, encode=False
            )
            cur_fixations = None
            for group_ids, group_X in groups:
                if copy_x is None:
                    copy_x = group_X[self.x]
                    copy_y = group_X[self.y]
                else:
                    copy_x = pd.concat(
                        [copy_x, group_X[self.x]], ignore_index=True, axis=0
                    )
                    copy_y = pd.concat(
                        [copy_y, group_X[self.y]], ignore_index=True, axis=0
                    )
                group_X[self.x] -= group_X[self.x].mean()
                group_X[self.y] -= group_X[self.y].mean()
                if cur_fixations is None:
                    cur_fixations = group_X
                else:
                    cur_fixations = pd.concat(
                        [cur_fixations, group_X], ignore_index=True, axis=0
                    )

            # === Correction of the AOI labels ===
            # Make the new aoi labels
            all_areas = np.unique(cur_fixations[self.aoi].values)
            if (self.n_aoi > 0) and (len(all_areas) > self.n_aoi):
                centers = []
                for i in range(len(all_areas)):
                    x = cur_fixations[(cur_fixations[self.aoi] == all_areas[i])][
                        self.x
                    ].mean()
                    y = cur_fixations[(cur_fixations[self.aoi] == all_areas[i])][
                        self.y
                    ].mean()
                    count = (
                        cur_fixations[(cur_fixations[self.aoi] == all_areas[i])]
                        .count()
                        .values[0]
                    )
                    centers.append([all_areas[i], count, x, y, True])
                count_of_aoi = len(all_areas)
                while count_of_aoi > self.n_aoi:
                    min_dist = np.inf
                    points_to_merge = []
                    for i in range(len(centers)):
                        for j in range(i + 1, len(centers)):
                            dist = np.linalg.norm(
                                np.array(centers[i][2:]) - np.array(centers[j][2:])
                            )
                            if (dist < min_dist) and centers[j][-1]:
                                min_dist = dist
                                points_to_merge = [centers[i], centers[j]]
                    cur_fixations.loc[
                        cur_fixations[self.aoi] == points_to_merge[1][0],
                        cur_fixations.columns == self.aoi,
                    ] = points_to_merge[0][0]
                    for i in range(len(centers)):
                        if centers[i][0] == points_to_merge[0][0]:
                            centers[i][2] = (
                                (points_to_merge[0][2] * points_to_merge[0][1])
                                + (points_to_merge[1][2] * points_to_merge[1][1])
                            ) / (points_to_merge[0][1] + points_to_merge[1][1])
                            centers[i][3] = (
                                (points_to_merge[0][3] * points_to_merge[0][1])
                                + (points_to_merge[1][3] * points_to_merge[1][1])
                            ) / (points_to_merge[0][1] + points_to_merge[1][1])
                            centers[i][1] = (
                                points_to_merge[0][1] + points_to_merge[1][1]
                            )
                        if centers[i][0] == points_to_merge[1][0]:
                            centers[i][-1] = False
                    count_of_aoi -= 1

            all_areas = np.unique(cur_fixations[self.aoi].values)
            areas_names = [f"aoi_{i}" for i in range(len(all_areas))]
            map_areas = dict(zip(all_areas, areas_names))
            cur_fixations[self.aoi] = cur_fixations[self.aoi].map(map_areas)
            used = []
            to_zip = []
            new_pattern = dict()
            # Find the sample with the same or less count of areas of interest to match the aoi labels
            for i in range(len(all_areas), 0, -1):
                if prev_pattern.get(i, 0) != 0:
                    pattern = prev_pattern[i]
                    # Compare areas
                    # Beginning with python 3.6, the order of dict.items() corresponds to the order which
                    # the elements are inserted
                    for key, value in pattern.items():
                        # if key not in used:
                        x_max, y_max, x_min, y_min = (
                            value[0],
                            value[1],
                            value[2],
                            value[3],
                        )
                        intersection = -1
                        new_name = None
                        prev_height = -1
                        prev_width = -1
                        for cur_area in areas_names:
                            cur_x_max, cur_y_max, cur_x_min, cur_y_min = (
                                cur_fixations[cur_fixations[self.aoi] == cur_area][
                                    self.x
                                ].max(),
                                cur_fixations[cur_fixations[self.aoi] == cur_area][
                                    self.y
                                ].max(),
                                cur_fixations[cur_fixations[self.aoi] == cur_area][
                                    self.x
                                ].min(),
                                cur_fixations[cur_fixations[self.aoi] == cur_area][
                                    self.y
                                ].min(),
                            )
                            width = min(x_max, cur_x_max) - max(x_min, cur_x_min)
                            height = min(y_max, cur_y_max) - max(y_min, cur_y_min)
                            # Find aoi with the largest intersection
                            if (
                                height > 0
                                and width > 0
                                and (cur_area not in used)
                                and intersection <= width * height
                            ):
                                intersection = width * height
                                new_name = cur_area
                                # prev_width = x_max - x_min
                                # prev_height = y_max - y_min
                        used.append(new_name)
                        to_zip.append(key)
                    len_of_used = len(used)
                    for j in range(len(areas_names) - len(used)):
                        used.append(None)
                        to_zip.append(areas_names[len_of_used + j])
                    for j in range(len(used)):
                        if used[j] is None:
                            for area in areas_names:
                                if area not in used:
                                    used[j] = area
                    # Match the remaining areas of interest
                    for j in range(len(used)):
                        if used[j] is None:
                            for k in range(len(areas_names)):
                                if areas_names[k] not in used:
                                    used[j] = areas_names[k]
                                    break
                    cur_fixations[self.aoi] = cur_fixations[self.aoi].map(
                        dict(zip(used, to_zip))
                    )
                    break
            # Add new sample
            for area in areas_names:
                cur_x_max, cur_y_max, cur_x_min, cur_y_min = (
                    cur_fixations[cur_fixations[self.aoi] == area][self.x].max(),
                    cur_fixations[cur_fixations[self.aoi] == area][self.y].max(),
                    cur_fixations[cur_fixations[self.aoi] == area][self.x].min(),
                    cur_fixations[cur_fixations[self.aoi] == area][self.y].min(),
                )
                new_pattern[area] = [cur_x_max, cur_y_max, cur_x_min, cur_y_min]
            # === End of the correction ===
            prev_pattern[len(np.unique(cur_fixations[self.aoi].values))] = new_pattern
            cur_fixations[self.x] = copy_x
            cur_fixations[self.y] = copy_y
            if fixations is None:
                fixations = cur_fixations
            else:
                fixations = pd.concat(
                    [fixations, cur_fixations], ignore_index=True, axis=0
                )

        return fixations



import io
from typing import Dict, List, Tuple, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy.typing import NDArray
from PIL import Image, ImageOps
from scipy.spatial import ConvexHull
from sklearn.base import BaseEstimator, TransformerMixin
from tqdm import tqdm

from eyefeatures.utils import _select_regressions, _split_dataframe


def _cmap_generation(n: int):
    return plt.cm.get_cmap("tab20")(n)


def scanpath_visualization(
    data_: pd.DataFrame,
    x: str,
    y: str,
    duration: str = None,
    dispersion: str = None,
    size_column: str = None,
    shape_column: str = None,
    time_stamps: str = None,
    aoi: str = None,
    img_path: str = None,
    fig_size: tuple[float, float] = (10.0, 10.0),
    points_width: float = 75,
    path_width: float = 0.004,
    points_color: str = "blue",
    path_color: str = "green",
    points_enumeration: bool = False,
    add_regressions: bool = False,
    regression_color: str = "red",
    micro_sac_color: str = None,
    is_vectors: bool = False,
    min_dispersion: float = 1.2,
    max_velocity: float = 4.7,
    aoi_c: Dict[str, str] = None,
    only_points: bool = False,
    seq_colormap: bool = False,
    show_hull: bool = False,
    show_legend: bool = False,
    path_to_img: str = None,
    with_axes: bool = False,
    axes_limits: tuple = None,
    rule: Tuple[int, ...] = None,
    deviation: Union[int, Tuple[int, ...]] = None,
    return_ndarray: bool = False,
    show_plot: bool = True,
    is_gray: bool = False,
    dpi: float = 100.0,
):
    """
    Function for scanpath visualization and aoi visualization.

    :param data_: DataFrame with fixations.
    :param x: x coordinate of fixation.
    :param y: y coordinate of fixation.
    :param duration: duration of fixation.
    :param dispersion: dispersion of fixation.
    :param size_column: label of the column, which is responsible for the size of the fixations(points on plot).\n
                        It can be duration, dispersion, etc.
    :param shape_column: label of the column, which is responsible for the shape of the fixations(points on plot).
    :param time_stamps: time stamps of fixations.
    :param aoi: AOI of fixations.
    :param img_path: path to the background image.
    :param fig_size: size of plot.
    :param points_width: width of points.
    :param path_width: width of path.
    :param points_color: color of points.
    :param path_color: color of path.
    :param points_enumeration: whether to enumerate points.
    :param add_regressions: whether to add regressions.
    :param regression_color: color of regressions.
    :param micro_sac_color: color of microsaccades.
    :param is_vectors: whether to visualize saccades as vectors
    :param min_dispersion: minimum dispersion.
    :param max_velocity: maximum velocity.
    :param aoi_c: colormap for AOI.
    :param only_points: whether to only show points.
    :param seq_colormap: whether to show sequentially-colored saccades.
    :param show_hull: whether to show hull of AOI.
    :param show_legend: whether to show legend.
    :param path_to_img: path to save the plot image.
    :param with_axes: whether to show axes.
    :param axes_limits: limits of axes.
    :param rule: must be either 1) tuple of quadrants direction to classify
                regressions, 1st quadrant being upper-right square of plane and counting
                anti-clockwise or 2) tuple of angles in degrees (0 <= angle <= 360).
    :param deviation: if None, then `rule` is interpreted as quadrants. Otherwise,
                    `rule` is interpreted as angles. If integer, then is a +-deviation for all angles.
                    If tuple of integers, then must be of the same length as `rule`, each value being
                    a corresponding deviation for each angle. Angle = 0 is positive x-axis direction,
                    rotating anti-clockwise.
    :param return_ndarray: whether to return numpy array of the plot image(returns RGBA array).
    :param show_plot: whether to show the plot.
    :param is_gray: whether to use the gray scale.
    """
    plt.figure(figsize=fig_size)
    eps = 1e-8
    marks = ("o", "^", "s", "*", "p")
    legend = dict()
    data = data_.copy()
    data["color"] = points_color
    if aoi is not None:
        data.dropna(subset=[aoi], inplace=True)

    data.reset_index(inplace=True, drop=True)
    X, Y = data[x], data[y]
    dX, dY = data[x].diff(), data[y].diff()
    XY = pd.concat([dX, dY], axis=1)
    fixation_size = np.full(X.shape[0], points_width)

    if shape_column is not None:
        intervals = np.linspace(0, data[shape_column].max(), 6)
        for i in range(1, len(intervals)):
            legend[marks[i - 1]] = (
                f"[{round(intervals[i - 1], 2)}, {round(intervals[i], 2)})"
            )
            data.loc[
                (data[shape_column] >= intervals[i - 1])
                & (data[shape_column] < intervals[i]),
                "mark",
            ] = marks[i - 1]
        markers = marks
    else:
        markers = [marks[0]]
        data["mark"] = marks[0]

    if aoi is not None:
        if aoi_c is None:
            get_aoi_cm = plt.cm.get_cmap(lut=len(data[aoi].drop_duplicates().values))
            aoi_c = dict()
            for i in range(len(data[aoi].unique())):
                aoi_c[data[aoi].unique()[i]] = get_aoi_cm(i)
        data["color"] = data[aoi].map(aoi_c)
    # print(data["color"])
    if size_column is not None:  # Not used
        size = data[size_column]
        size /= size.max()
        fixation_size = np.array(size * points_width)

    plt.axis(axes_limits)

    if img_path is not None:
        plt.imshow(plt.imread(img_path))

    for i in range(len(markers)):
        if shape_column is not None:
            plt.scatter(
                x=data[x][data["mark"] == markers[i]],
                y=data[y][data["mark"] == markers[i]],
                color=data["color"][data["mark"] == markers[i]],
                marker=markers[i],
                edgecolors="black",
                label=legend[marks[i - 1]],
            )
        else:
            plt.scatter(
                x=data[x][data["mark"] == markers[i]],
                y=data[y][data["mark"] == markers[i]],
                color=data["color"][data["mark"] == markers[i]],
                marker=markers[i],
                edgecolors="black",
            )

    if aoi is not None and show_hull:
        for area in data[aoi].drop_duplicates().values:
            points = data[data[aoi] == area]
            assert points[x].shape[0] > 2, "Error: Need more points for aoi"
            points_num = np.array([points[x].to_numpy(), points[y].to_numpy()])
            points_num = points_num.T

            hull = ConvexHull(points_num)
            for simplex in hull.simplices:
                plt.plot(
                    points_num[simplex, 0],
                    points_num[simplex, 1],
                    "-",
                    color=aoi_c[area],
                )
            plt.fill(
                points_num[hull.vertices, 0],
                points_num[hull.vertices, 1],
                color=aoi_c[area],
                alpha=0.5,
                label=area,
            )
            # legend[area] = area
    elif aoi is not None and not show_hull:
        for area in data[aoi].drop_duplicates().values:
            plt.fill(
                data.loc[(data[aoi] == area), (data.columns.isin([x]))].values[0],
                data.loc[(data[aoi] == area), (data.columns.isin([y]))].values[0],
                color=aoi_c[area],
                alpha=0.5,
                label=area,
            )

    if show_legend:
        plt.legend()

    if points_enumeration:
        enumeration = range(dX.shape[0])
        for i in enumeration:
            plt.annotate(str(i), xy=(X[i], Y[i]))

    if not only_points:
        c_sac = np.array(mpl.colors.to_rgb(path_color))
        if seq_colormap:
            ls = np.linspace(0, 1, X.shape[0])
        else:
            ls = np.ones(X.shape[0])

        if not is_vectors:
            for i in range(len(X) - 1):
                plt.plot(
                    [X.iloc[i], X.iloc[i + 1]],
                    [Y.iloc[i], Y.iloc[i + 1]],
                    color=c_sac * ls[i],
                    linewidth=path_width,
                )
            if add_regressions:
                mask = _select_regressions(dX, dY, rule, deviation)
                c_reg = np.array(mpl.colors.to_rgb(regression_color))
                regX = dX[mask]
                for i in regX.index:
                    if i != 0:
                        plt.plot(
                            [X.iloc[i - 1], X.iloc[i]],
                            [Y.iloc[i - 1], Y.iloc[i]],
                            color=c_reg * ls[i],
                            linewidth=path_width,
                        )
        else:
            for i in range(len(X) - 1):
                plt.quiver(
                    X.iloc[i],
                    Y.iloc[i],
                    X.iloc[i + 1] - X.iloc[i],
                    Y.iloc[i + 1] - Y.iloc[i],
                    angles="xy",
                    scale_units="xy",
                    color=c_sac * ls[i],
                    scale=1,
                    width=path_width,
                    edgecolor="yellow",
                    linewidth=path_width / 2,
                )
            if add_regressions:
                mask = _select_regressions(dX, dY, rule, deviation)

                c_reg = np.array(mpl.colors.to_rgb(regression_color))
                regX = dX[mask]
                for i in regX.index:
                    if i != 0:
                        plt.quiver(
                            X.iloc[i - 1],
                            Y.iloc[i - 1],
                            X.iloc[i] - X.iloc[i - 1],
                            Y.iloc[i] - Y.iloc[i - 1],
                            angles="xy",
                            scale_units="xy",
                            color=c_reg * ls[i],
                            scale=1,
                            width=path_width,
                            edgecolor="yellow",
                            linewidth=path_width / 2,
                        )

        if (
            micro_sac_color is not None
        ):  # TODO (doesn't work with normalized coordinates)
            assert (
                dispersion is not None
            ), "Error: provide 'dispersion' column before calling visualization"
            assert (
                time_stamps is not None
            ), "Error: provide 'time_stamps' column before calling visualization"
            assert (
                duration is not None
            ), "Error: provide 'duration' column before calling visualization"
            dr = np.sqrt(dX**2 + dY**2)
            dt = data[time_stamps] - (data[time_stamps] + data[duration] / 1000).shift(
                1
            )
            v = dr / (dt + eps)

            mic_sac = data[(data[dispersion] > min_dispersion) & (v < max_velocity)]
            mic_sacX = np.array([(X.iloc[i - 1], X.iloc[i]) for i in mic_sac.index])
            mic_sacY = np.array([(Y.iloc[i - 1], Y.iloc[i]) for i in mic_sac.index])
            for i in range(len(mic_sacX)):
                plt.plot(
                    mic_sacX[i],
                    mic_sacY[i],
                    color=micro_sac_color,
                    linewidth=path_width,
                )

    if not with_axes:
        plt.axis("off")

    arr = None

    if return_ndarray:
        image_buffer = io.BytesIO()
        plt.savefig(image_buffer, dpi=dpi, format="png")
        im = Image.open(image_buffer).convert("RGB")
        if is_gray:
            im = ImageOps.grayscale(im)
        arr = np.array(im) / 255
        image_buffer.close()

    if path_to_img is not None:
        plt.savefig(path_to_img, dpi=dpi)
        plt.axis("on")
    if not show_plot:
        plt.close()
    return arr


class Visualization(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        x: str,
        y: str,
        duration: str = None,
        dispersion: str = None,
        size_column: str = None,
        shape_column: str = None,
        time_stamps: str = None,
        aoi: str = None,
        img_path: str = None,
        fig_size: tuple[float, float] = (10.0, 10.0),
        points_width: float = 75,
        path_width: float = 0.004,
        points_color: str = "blue",
        path_color: str = "green",
        points_enumeration: bool = False,
        add_regressions: bool = False,
        regression_color: str = "red",
        micro_sac_color: str = None,
        is_vectors: bool = False,
        min_dispersion: float = 1.2,
        max_velocity: float = 4.7,
        aoi_c: Dict[str, str] = None,
        only_points: bool = False,
        seq_colormap: bool = False,
        show_hull: bool = False,
        show_legend: bool = False,
        path_to_img: str = None,
        with_axes: bool = False,
        axes_limits: tuple = None,
        rule: Tuple[int, ...] = None,
        deviation: Union[int, Tuple[int, ...]] = None,
        is_gray: bool = False,
    ):
        self.x = x
        self.y = y
        self.duration = duration
        self.dispersion = dispersion
        self.size_column = size_column
        self.shape_column = shape_column
        self.time_stamps = time_stamps
        self.aoi = aoi
        self.img_path = img_path
        self.fig_size = fig_size
        self.points_width = points_width
        self.points_color = points_color
        self.path_color = path_color
        self.add_regressions = add_regressions
        self.regression_color = regression_color
        self.micro_sac_color = micro_sac_color
        self.is_vectors = is_vectors
        self.min_dispersion = min_dispersion
        self.max_velocity = max_velocity
        self.show_legend = show_legend
        self.path_to_img = path_to_img
        self.path_width = path_width
        self.points_enumeration = points_enumeration
        self.show_hull = show_hull
        self.aoi_c = aoi_c
        self.only_points = only_points
        self.seq_colormap = seq_colormap
        self.with_axes = with_axes
        self.axes_limits = axes_limits
        self.rule = rule
        self.deviation = deviation
        self.return_ndarray = True
        self.show_plot = False
        self.is_gray = is_gray

    def fit(self, X: pd.DataFrame, y=None):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        return scanpath_visualization(data_=X, **self.__dict__)


def get_visualizations(
    data: pd.DataFrame,
    x: str,
    y: str,
    shape: tuple[int, int],
    pattern: str,
    dpi: float = 100.0,
    pk: List[str] = None,
):
    """
    Gets visualizations.

    :param data: Dataframe with fixations
    :param patterns: List of Visualizations
    :param pk: list of column names used to split pd.DataFrame.

    Returns
    -------
    output: (ndarray)
        [n, m, fig, fig, c]\n
         n   : instances\n
         m   : patterns\n
         fig : fig size\n
         c   : (3 - RGB image; 1 - gray scaled image)
    """
    arr = []
    if pk is None:
        res = []
        # for pattern in patterns:
        #     res.append(pattern.transform(data))
        if pattern == "baseline":
            res = baseline_visualization(data, x, y, shape)
        elif pattern == "aoi":
            res = aoi_visualization(data, x, y, shape, aoi="AOI")
        elif pattern == "saccades":
            res = saccade_visualization(data, x, y, shape)
        else:
            raise ValueError(f"Unsupported pattern: {pattern}")
        arr.append(res)
    else:
        groups: List[str, pd.DataFrame] = _split_dataframe(data, pk, encode=False)
        for group_ids, group_X in tqdm(groups):
            cur_data = data[pd.DataFrame(data[pk] == group_ids).all(axis=1)]
            if pattern == "baseline":
                res = baseline_visualization(
                    cur_data, x, y, shape, show_plot=False, dpi=dpi
                )
            elif pattern == "aoi":
                res = aoi_visualization(
                    cur_data, x, y, shape, aoi="AOI", show_plot=False, dpi=dpi
                )
            elif pattern == "saccades":
                res = saccade_visualization(
                    cur_data, x, y, shape, show_plot=False, dpi=dpi
                )
            else:
                raise ValueError(f"Unsupported pattern: {pattern}")
            arr.append(res)
    arr = np.transpose(np.array(arr), (0, 3, 1, 2))
    print(arr.shape)
    return arr


def baseline_visualization(
    data_: pd.DataFrame,
    x: str,
    y: str,
    shape: tuple[int, int] = (10, 10),
    path_width: float = 1,
    show_legend: bool = False,
    path_to_img: str = None,
    with_axes: bool = False,
    show_plot: bool = False,
    return_ndarray: bool = True,
    dpi: float = 100.0,
):
    return scanpath_visualization(
        data_,
        x,
        y,
        fig_size=shape,
        show_legend=show_legend,
        path_to_img=path_to_img,
        with_axes=with_axes,
        path_width=path_width,
        return_ndarray=return_ndarray,
        show_plot=show_plot,
        dpi=dpi,
    )


def aoi_visualization(
    data_: pd.DataFrame,
    x: str,
    y: str,
    shape: tuple[int, int] = (10, 10),
    aoi: str = "AOI",
    size_column: str = None,
    shape_column: str = None,
    img_path: str = None,
    points_width: float = 75,
    path_width: float = 1,
    points_color: str = None,
    aoi_c: Dict[str, str] = None,
    seq_colormap: bool = False,
    show_legend: bool = False,
    path_to_img: str = None,
    with_axes: bool = False,
    axes_limits: tuple = None,
    return_ndarray: bool = True,
    show_plot: bool = True,
    only_points: bool = True,
    dpi: float = 100.0,
):
    return scanpath_visualization(
        data_,
        x,
        y,
        size_column=size_column,
        shape_column=shape_column,
        aoi=aoi,
        img_path=img_path,
        fig_size=shape,
        points_width=points_width,
        path_width=path_width,
        points_color=points_color,
        seq_colormap=seq_colormap,
        show_legend=show_legend,
        aoi_c=aoi_c,
        with_axes=with_axes,
        axes_limits=axes_limits,
        path_to_img=path_to_img,
        show_hull=True,
        only_points=only_points,
        return_ndarray=return_ndarray,
        show_plot=show_plot,
        dpi=dpi,
    )


def saccade_visualization(
    data_: pd.DataFrame,
    x: str,
    y: str,
    shape: tuple[int, int] = (10, 10),
    size_column: str = None,
    shape_column: str = None,
    duration: str = None,
    dispersion: str = None,
    time_stamps: str = None,
    img_path: str = None,
    path_width: float = 1,
    path_color: str = "green",
    add_regressions: bool = False,
    regression_color: str = "red",
    micro_sac_color: str = None,
    is_vectors: bool = False,
    min_dispersion: float = 1.2,
    max_velocity: float = 4.7,
    seq_colormap: bool = False,
    show_legend: bool = False,
    path_to_img: str = None,
    with_axes: bool = False,
    axes_limits: tuple = None,
    rule: Tuple[int, ...] = (2,),
    deviation: Union[int, Tuple[int, ...]] = None,
    return_ndarray: bool = True,
    show_plot: bool = True,
    dpi: float = 100.0,
):
    return scanpath_visualization(
        data_,
        x,
        y,
        size_column=size_column,
        shape_column=shape_column,
        duration=duration,
        dispersion=dispersion,
        time_stamps=time_stamps,
        img_path=img_path,
        fig_size=shape,
        path_width=path_width,
        seq_colormap=seq_colormap,
        show_legend=show_legend,
        path_to_img=path_to_img,
        with_axes=with_axes,
        axes_limits=axes_limits,
        rule=rule,
        deviation=deviation,
        path_color=path_color,
        regression_color=regression_color,
        add_regressions=add_regressions,
        micro_sac_color=micro_sac_color,
        is_vectors=is_vectors,
        min_dispersion=min_dispersion,
        max_velocity=max_velocity,
        return_ndarray=return_ndarray,
        show_plot=show_plot,
        dpi=dpi,
    )



from typing import Any, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from numba import jit
from sklearn.base import BaseEstimator, TransformerMixin


class BaseTransformer(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        x: str = None,
        y: str = None,
        t: str = None,
        duration: str = None,
        dispersion: str = None,
        aoi: str = None,
        path_pk: List[str] = None,
        pk: List[str] = None,
        expected_paths: Dict[str, pd.DataFrame] = None,
        fill_path: pd.DataFrame = None,
        expected_paths_method: str = "mean",
        return_df: bool = True,
    ):
        self.x = x
        self.y = y
        self.t = t
        self.duration = duration
        self.dispersion = dispersion
        self.path_pk = path_pk
        self.pk = pk
        self.aoi = aoi
        self.return_df = return_df
        self.expected_paths = expected_paths
        self.fill_path = fill_path
        self.expected_paths_method = expected_paths_method

    def _check_init(self, items: List[Tuple[Any, str]]):
        for value, nm in items:
            if value is None:
                raise RuntimeError(f"{nm} is not initialized")

    def set_data(
        self,
        x: str = None,
        y: str = None,
        t: str = None,
        duration: str = None,
        dispersion: str = None,
        aoi: str = None,
        path_pk: List[str] = None,
        pk: List[str] = None,
        expected_paths: Dict[str, pd.DataFrame] = None,
        fill_path: pd.DataFrame = None,
        expected_paths_method: str = "mean",
        return_df: bool = True,
    ):
        self.x = x
        self.y = y
        self.t = t
        self.duration = duration
        self.dispersion = dispersion
        self.path_pk = path_pk
        self.pk = pk
        self.aoi = aoi
        self.return_df = return_df
        self.expected_paths = expected_paths
        self.fill_path = fill_path
        self.expected_paths_method = expected_paths_method

    def fit(self, X: pd.DataFrame, y=None):
        return self

    def transform(self, X: pd.DataFrame) -> Union[pd.DataFrame, np.ndarray]:
        return X if self.return_df else X.values


class Extractor(BaseEstimator, TransformerMixin):  # TODO rename to FeatureExtractor
    def __init__(
        self,
        features: List[BaseTransformer] = None,
        x: str = None,
        y: str = None,
        t: str = None,
        duration: str = None,
        dispersion: str = None,
        aoi: str = None,
        path_pk: List[str] = None,
        pk: List[str] = None,
        expected_paths_method: str = "mean",
        extra: List[str] = None,
        aggr_extra: str = None,
        return_df: bool = True,
    ):
        self.features = features
        self.x = x
        self.y = y
        self.t = t
        self.duration = duration
        self.dispersion = dispersion
        self.aoi = aoi
        self.path_pk = path_pk
        self.pk = pk
        self.expected_paths_method = expected_paths_method
        self.extra = extra
        self.aggr_extra = aggr_extra
        self.return_df = return_df
        self.is_fitted = False

    @jit(forceobj=True, looplift=True)
    def fit(self, X: pd.DataFrame, y=None):
        self.is_fitted = True
        if self.features is not None:
            for feature in self.features:
                feature.set_data(
                    x=self.x,
                    y=self.y,
                    t=self.t,
                    duration=self.duration,
                    dispersion=self.dispersion,
                    aoi=self.aoi,
                    path_pk=self.path_pk,
                    pk=self.pk,
                    expected_paths_method=self.expected_paths_method,
                    return_df=self.return_df,
                )
                feature.fit(X)

        return self

    @jit(forceobj=True, looplift=True)
    def transform(self, X: pd.DataFrame) -> Union[pd.DataFrame, np.ndarray]:
        if not self.is_fitted:
            raise RuntimeError("Class is not fitted")

        if self.features is None:
            return X if self.return_df else X.values

        gathered_features = []
        data_df: pd.DataFrame = X[
            [self.x, self.y, self.t, self.duration, self.dispersion]
        ]

        if self.pk is not None:
            data_df = pd.concat([data_df, X[self.pk]], axis=1)

        if self.aoi is not None:
            data_df = pd.concat([data_df, X[self.aoi]], axis=1)

        for feature in self.features:
            gathered_features.append(feature.transform(data_df))

        if self.extra is not None:
            gathered_features.append(
                X[self.extra].groupby(self.pk).apply(self.aggr_extra)
            )

        features_df = pd.concat(gathered_features, axis=1)

        return features_df if self.return_df else features_df.values



from typing import Callable, Dict, List, Union

import numpy as np
import pandas as pd
from numba import jit
from numpy.typing import NDArray
from scipy.cluster.hierarchy import leaves_list, linkage, optimal_leaf_ordering
from scipy.linalg import sqrtm
from scipy.optimize import minimize
from scipy.sparse.csgraph import laplacian
from scipy.sparse.linalg import eigsh
from sklearn.manifold import MDS
from tqdm import tqdm

from eyefeatures.utils import Types, _split_dataframe


def _target_norm(fwp: np.ndarray, fixations: np.ndarray) -> float:
    return np.linalg.norm(fixations - fwp, axis=1).sum()


def get_expected_path(
    data: pd.DataFrame,
    x: str,
    y: str,
    path_pk: List[str],
    pk: List[str],
    method: str = "mean",
    duration: str = None,
    return_df: bool = True,
) -> Dict[str, Union[pd.DataFrame, np.ndarray]]:
    """
    Estimates expected path by a given method

    :param data: pd.Dataframe containing coordinates of fixations and its timestamps
    :param x: Column name of x-coordinate
    :param y: Column name of y-coordinate
    :param path_pk: List of column names of groups to calculate expected path (must be a subset of pk)
    :param pk: List of column names used to split pd.Dataframe
    :param method: method to calculate expected path ("mean" or "fwp")
    :param duration: Column name of fixations duration if needed
    :param return_df: Return pd.Dataframe object else np.ndarray
    :return: Dict of groups and Union[pd.Dataframe, np.ndarray] of form (x_est, y_est) or (x_est, y_est, duration_est)
    """

    if not set(path_pk).issubset(set(pk)):
        raise ValueError("path_pk must be a subset of pk")

    columns = [x, y]
    columns_ret = ["x_est", "y_est"]
    if duration is not None:
        columns.append(duration)
        columns_ret.append("duration_est")

    resulted_paths = {}
    path_groups: Types.Partition = _split_dataframe(data, path_pk)

    for group_nm, group_path in path_groups:
        expected_path = []
        data_part: Types.Partition = _split_dataframe(group_path, pk)
        path_length = max(len(part_path) for part_nm, part_path in data_part)

        if method == "mean":
            reshaped_paths = [
                np.pad(
                    part_path[columns],
                    pad_width=((0, path_length - len(part_path)), (0, 0)),
                    mode="constant",
                )
                for part_nm, part_path in data_part
            ]
            expected_path = np.mean(np.array(reshaped_paths), axis=0)
        elif method == "fwp":
            for step in range(path_length):
                duration_sum = 0
                observed_points = []
                for part_nm, part_path in data_part:
                    part_path = part_path[columns].values
                    if part_path.shape[0] > step:
                        observed_points.append(part_path[step, :2])
                        if len(columns) == 3:
                            duration_sum += part_path[step, 2]

                vector_points = np.array(observed_points)
                fwp_init = np.mean(vector_points, axis=0)
                fwp_opt = minimize(
                    _target_norm, fwp_init, args=(vector_points,), method="L-BFGS-B"
                )
                fix_opt = [fwp_opt.x[0], fwp_opt.x[1]]
                if len(columns) == 3:
                    fix_opt.append(duration_sum / max(1, len(observed_points)))

                expected_path.append(fix_opt)
        else:
            raise ValueError('Only "mean" and "fwp" methods are supported')

        expected_path_df = pd.DataFrame(expected_path, columns=columns_ret)
        resulted_paths[group_nm] = (
            expected_path_df if return_df else expected_path_df.values
        )

    return resulted_paths


def _get_fill_path(
    data: List[pd.DataFrame],
    x: str,
    y: str,
    method: str = "mean",
    duration: str = None,
) -> pd.DataFrame:
    """
    Calculates fill path as expected path of given expected paths

    :param data: paths data
    :param x: Column name of x-coordinate
    :param y: Column name of y-coordinate
    :param method: method to calculate expected path ("mean" or "fwp")
    :param duration: Column name of fixations duration if needed
    """

    paths = pd.concat(
        [path.assign(dummy=k) for k, path in enumerate(data)], ignore_index=True
    )

    fill_path = get_expected_path(
        data=paths,
        x=x,
        y=y,
        path_pk=["dummy"],
        pk=["dummy"],
        method=method,
        duration=duration,
    )
    return list(fill_path.values())[0]


# ======================== SIMILARITY MATRIX ========================
def restore_matrix(matrix: NDArray, tol=1e-9):
    """
    Estimates A assuming 'matrix' equals
    .. math: $A^T A$.
    """
    # Получаем собственные вектора и диагональную матрицу с.ч.
    evals, evecs = np.linalg.eigh(matrix)

    # Сортируем вектора и числа по убыванию модуля
    sorted_evals_indexes = np.argsort(evals)[::-1]
    # sorted_evals_indexes = np.argsort(np.abs(evals))[::-1]
    evecs = evecs[:, sorted_evals_indexes]
    evals = evals[sorted_evals_indexes]

    # Удалим вырожденные части из матриц
    A_rank = (evals > tol).sum()
    # A_rank = (np.abs(evals) > tol).sum()
    evals = np.diag(evals[:A_rank])
    evecs = evecs[:, :A_rank]

    # Посчитам матрицу A
    S = sqrtm(evals)
    U = np.identity(A_rank)

    US = U.dot(S)
    return US.dot(evecs.T).T, A_rank


def get_sim_matrix(scanpaths: List[NDArray], sim_metric: Callable) -> np.ndarray:
    """
    Computes similarity matrix given non-trivial metric.

    :param scanpaths: list of scanpaths, each being 2D-array of shape (2, n).
    :param sim_metric: similarity metric.
    :return: scaled similarity matrix.
    """
    n = len(scanpaths)
    sim_matrix = np.ones(shape=(n, n))
    for i in range(len(scanpaths)):
        for j in range(i + 1, len(scanpaths)):
            p, q = scanpaths[i], scanpaths[j]
            sim_matrix[i, j] = sim_metric(p, q)

    sim_matrix += sim_matrix.T
    m = np.max(sim_matrix)
    m = m if m != 0 else 1
    return sim_matrix / m


@jit(forceobj=True, looplift=True)
def get_dist_matrix(
    scanpaths: List[pd.DataFrame], dist_metric: Callable
) -> pd.DataFrame:
    """
    Computes pairwise distance matrix given distance metric.

    :param scanpaths: List of scanpaths DataFrames of form (x, y)
    :param dist_metric: Metric used to calculate distance from features.scanpath_dist
    """

    if len(scanpaths) == 0:
        raise ValueError("scanpaths list is empty")

    distances = []
    for i in tqdm(range(len(scanpaths))):
        for j in range(i + 1):
            distance = dist_metric(scanpaths[i], scanpaths[j])
            distances.append([i, j, distance])
            distances.append([j, i, distance])

    distances_df = pd.DataFrame(distances, columns=["p", "q", "dist"])
    return distances_df.reset_index().pivot_table(index="p", columns="q", values="dist")


def hierarchical_clustering_order(
    sim_matrix: np.ndarray, metric: str = "euclidean"
) -> np.ndarray:
    """
    Reorder matrix using hierarchical clustering.

    :param sim_matrix: similarity matrix to reorder
    :param metric: metric used in building matrix
    :return: reordered matrix
    """

    Z = linkage(sim_matrix, method="ward", metric=metric)
    ordered_leaves = leaves_list(Z)
    reordered_matrix = sim_matrix[ordered_leaves, :][:, ordered_leaves]
    return reordered_matrix


def optimal_leaf_ordering_clustering(
    sim_matrix: np.ndarray, metric: str = "euclidean"
) -> np.ndarray:
    """
    Reorder matrix using optimal leaf ordering.

    :param sim_matrix: similarity matrix to reorder
    :return: reordered matrix
    """

    Z = linkage(sim_matrix, method="ward", metric=metric)
    Z = optimal_leaf_ordering(Z, sim_matrix)
    ordered_leaves = leaves_list(Z)
    reordered_matrix = sim_matrix[ordered_leaves, :][:, ordered_leaves]
    return reordered_matrix


def dimensionality_reduction_order(sim_matrix: np.ndarray) -> np.ndarray:
    """
    Reorder matrix using Multi-Dimensional Scaling (MDS).

    :param sim_matrix: similarity matrix to reorder
    :return: reordered matrix
    """

    mds = MDS(n_components=2, dissimilarity="precomputed")
    coords = mds.fit_transform(1 - sim_matrix)
    # reorder the matrix based on the first MDS component
    order = np.argsort(coords[:, 0])
    reordered_matrix = sim_matrix[order, :][:, order]
    return reordered_matrix


def spectral_order(sim_matrix: np.ndarray) -> np.ndarray:
    """
    Reorder matrix using spectral reordering.

    :param sim_matrix: similarity matrix to reorder
    :return: reordered matrix
    """

    L = laplacian(sim_matrix, normed=True)

    _, eigenvectors = eigsh(L, k=2, which="SM")

    # reorder based on the second smallest eigenvector (fiedler vector)
    fiedler_vector = eigenvectors[:, 1]
    order = np.argsort(fiedler_vector)
    reordered_matrix = sim_matrix[order, :][:, order]
    return reordered_matrix


# ======================== COMPROMISE MATRIX ========================


def get_center_matrix(weight_vector: np.ndarray) -> np.ndarray:
    """
    Calculates centering matrix Theta.

    :param weight_vector: vector of weights
    :return: centering matrix
    """
    assert np.sum(weight_vector) > 0, "Sum of weights must be greater than 0"
    weight_vector = weight_vector.astype(np.float32)
    weight_vector /= np.sum(weight_vector)
    E = np.eye(weight_vector.size)
    Theta = E - np.matmul(
        np.ones(weight_vector.size)[:, np.newaxis], weight_vector[:, np.newaxis].T
    )
    return Theta


def get_cross_product_matrix(
    D: np.ndarray, weight_vector: np.ndarray = None
) -> np.ndarray:
    """
    Calculates cross-product matrix.

    :param D: distance matrix
    :param weight_vector: vector of weights
    :return: cross-product matrix
    """

    if weight_vector is None:
        weight_vector = np.ones(D.shape[0])

    Theta = get_center_matrix(weight_vector)
    return -0.5 * Theta @ D @ Theta.T


def compute_rv_coefficient(S1: np.ndarray, S2: np.ndarray) -> float:
    """
    Calculate the RV coefficient between two cross-product matrices.

    :param S1: first cross-product matrix
    :param S2: second cross-product matrix
    :return: RV coefficient
    """
    numerator = np.trace(S1 @ S2.T)
    denominator = np.sqrt(np.trace(S1 @ S1.T) * np.trace(S2 @ S2.T))
    return numerator / denominator


def get_compromise_matrix(distance_matrices: List[np.ndarray]) -> np.ndarray:
    """
    Compute the compromise matrix from a list of distance matrices.

    :param distance_matrices: List of distance matrices (each an ndarray)
    :return: compromise cross-product matrix
    """
    assert len(distance_matrices) > 0, "At least one distance matrix is required"
    num_matrices = len(distance_matrices)

    # compute the cross-product matrices
    cross_product_matrices = [get_cross_product_matrix(D) for D in distance_matrices]

    # compute the similarity matrix using RV coefficients
    similarity_matrix = np.zeros((num_matrices, num_matrices))
    for i in range(num_matrices):
        for j in range(i, num_matrices):
            rv = compute_rv_coefficient(
                cross_product_matrices[i], cross_product_matrices[j]
            )
            similarity_matrix[i, j] = similarity_matrix[j, i] = rv

    # eigen-decomposition of the similarity matrix
    _, eigvecs = np.linalg.eigh(similarity_matrix)
    weights = eigvecs[:, -1]  # eigenvector corresponding to the largest eigenvalue
    # Get the compromise matrix as a weighted sum
    comp_mat = sum(w * G for w, G in zip(weights, cross_product_matrices))

    return comp_mat



from typing import Callable, List, Literal, Tuple, Union

import gudhi as gd
import numpy as np
import pandas as pd
from numba import jit, prange
from numpy.typing import NDArray
from PyEMD.EMD2d import EMD2D
from scipy.signal import convolve2d
from scipy.stats import gaussian_kde

from eyefeatures.utils import _rec2square, _split_dataframe


# =========================== HEATMAPS ===========================
def _check_shape(shape: Tuple[int, int]):
    assert isinstance(shape, tuple), f"'shape' must be tuple, hot {type(shape)}."
    assert len(shape) == 2, f"'shape' must be of length 2, got {len(shape)}."
    for k in shape:
        assert isinstance(
            k, int
        ), f"Values in 'shape' must be integers, got '{type(k)}'."
        assert k > 0, f"Integers in 'shape' must be positive, got '{k}'."


def get_heatmap(x: NDArray, y: NDArray, shape: Tuple[int, int], check: bool = True):
    """
    Get heatmap from scanpath (given coordinates are scaled and sorted in time) using
    Gaussian KDE.
    :param x: x coordinates of scanpath.
    :param y: y coordinates of scanpath.
    :param shape: if tuple with single integer, then square matrix is returned, otherwise k must be (height, width)
                  tuple and rectangular matrix is returned.
    :param check: whether to check 'shape' for correct typing.
    :return: heatmap matrix.
    """
    if check:
        _check_shape(shape)

    if len(x) <= 2:  # TODO warning
        # in case of small number of samples, KDE cannot be applied and
        # default kernel estimate is returned instead
        x, y = np.array([0.25, 0.50, 0.75]), np.array([0.50, 0.50, 0.50])

    scanpath = np.vstack([x, y])
    kernel = gaussian_kde(scanpath)
    interval_x, interval_y = np.linspace(0, 1, shape[1]), np.linspace(0, 1, shape[0])
    x, y = np.meshgrid(interval_x, interval_y)

    positions = np.vstack([y.ravel(), x.ravel()])
    return np.reshape(kernel(positions), x.shape)


def get_heatmaps(
    data: pd.DataFrame, x: str, y: str, shape: Tuple[int, int], pk: List[str] = None
):
    """
    Convenience wrapper for get_heatmap function.
    """
    _check_shape(shape)

    if pk is None:
        x_path, y_path = data[x].values, data[y].values
        heatmap = get_heatmap(x_path, y_path, shape)
        heatmaps = heatmap[np.newaxis, :, :]
    else:
        groups: List[str, pd.DataFrame] = _split_dataframe(data, pk)
        hshape = (len(groups), shape[0], shape[1])

        heatmaps = np.zeros(hshape)
        for i, (_, group_X) in enumerate(groups):
            x_path, y_path = group_X[x], group_X[y]
            heatmaps[i, :, :] = get_heatmap(x_path, y_path, shape, check=False)

    return heatmaps


# =========================== PCA ===========================
def pca(matrix: NDArray, p: int, cum_sum: float = None):
    """
    PCA compression.

    :param matrix: matrix to get principal components from (n x m)
    :param p: number of first principal components to leave
    :param cum_sum: instead of p, leave such number of principal components, that 0.0 <= a_cum_sum <= 1.0
                    fraction of information is conserved.
    :return: matrix of eigenvectors (m x m), projection (n x p), rows means (n x 1)
    """
    assert len(matrix.shape) == 2, "'matrix' should be a matrix"
    assert 0 <= p <= matrix.shape[0], "given 'matrix' is n x m, 0 <= p <= n must hold"
    assert (p is not None) or (
        cum_sum is not None
    ), "either 'p' or 'cum_sum' must be provided"
    assert (cum_sum is None) or (
        0.0 <= cum_sum <= 1.0
    ), "'cum_sum' must be between 0.0 and 1.0"

    matrix = matrix.astype(np.float64)

    row_means = np.mean(matrix, axis=1)
    matrix -= row_means[:, None]
    c = np.cov(matrix)
    evals, evecs = np.linalg.eigh(c)
    sorted_evals_indexes = np.argsort(evals)[::-1]
    evecs = evecs[:, sorted_evals_indexes]

    if p is not None:
        evecs = evecs[:, :p]
    else:
        evals /= evals.sum()
        cumsum = 0.0
        p = 0
        while cumsum < cum_sum and p < evals.size():
            cumsum += evals[p]
            p += 1

        evecs = evecs[:, :p]

    projection = evecs.T @ matrix
    return evecs, projection, row_means


# =========================== RQA ===========================
@jit(forceobj=True, looplift=True)
def get_rqa(
    data: pd.DataFrame, x: str, y: str, metric: Callable, rho: float
) -> np.ndarray:
    """
    Calculates recurrence quantification analysis matrix based on given fixations.
    :param data: dataframe containing fixations
    :param x: column name of x-coordinate
    :param y: column name of y-coordinate
    :param metric: callable metric on R^2 points
    :param rho: threshold radius

    :return: rqa matrix
    """
    fixations = data[[x, y]].values
    n = len(fixations)
    rqa_matrix = np.zeros((n, n), dtype=np.int32)

    for i in prange(n):
        for j in prange(i + 1, n):
            dist = metric(fixations[i], fixations[j])
            rqa_matrix[i][j] = int(dist <= rho)
            rqa_matrix[j][i] = int(dist <= rho)

    return rqa_matrix


# =========================== MTF ===========================
def get_mtf(
    data: pd.DataFrame,
    x: str,
    y: str,
    n_bins: int = 10,
    output_size: Union[int, float] = 1.0,
    shrink_strategy: Literal["max", "mean", "normal"] = "normal",
    flatten: bool = False,
) -> np.array:
    """
    Calculates Markov Transition Field for (x,y) coordinates.
    :param data: dataframe containing fixation coordinates.
    :param x: x-coordinate column name.
    :param y: y-coordinate column name.
    :param n_bins: number of bins to discretize time series into.
    :param output_size: fraction between 0 and 1. Specifies fraction of input series length to shrink output to.
    :param shrink_strategy: strategy to use for convolution while shrinking. Ignored if 'output_size' is equal to
                            size of 'data'.
    :param flatten: bool, whether to flatten the array.

    :returns: tensor of shape (2, n_coords, n_coords), where n_coords is the length of input dataframe.
    """
    if isinstance(output_size, float):
        assert 0.0 < output_size <= 1.0, "Must be 0 < output_size <= 1."
        output_size = np.ceil(output_size * len(data)).astype(np.int64)

    elif isinstance(output_size, int):
        assert (
            0 < output_size <= len(data)
        ), "'output_size' must be positive integer not exceeding the input size."

    else:
        raise ValueError(f"'output_size' is of wrong type '{type(output_size)}'.")

    assert n_bins > 1, "'n_bins' must be greater than 1."
    assert len(data) > n_bins, "Input series must contain more than 'n_bins' samples."

    x_coords, y_coords = data[[x]].values.ravel(), data[[y]].values.ravel()
    fixations_mtf = _get_mtf(np.array([x_coords, y_coords]), n_bins=n_bins)

    n_samples, n_timestamps = 2, len(x_coords)
    if output_size < n_timestamps:
        shrunk_mtfs = []
        for i in prange(n_samples):
            shrunk_mtfs.append(
                _shrink_matrix(
                    fixations_mtf[i, :, :],
                    height=output_size,
                    width=output_size,
                    strategy=shrink_strategy,
                )
            )

        del fixations_mtf
        fixations_mtf = np.array(shrunk_mtfs)

    return fixations_mtf.flatten() if flatten else fixations_mtf


@jit(forceobj=True, looplift=True)
def _get_mtf(a: np.array, n_bins: int) -> np.array:
    assert len(a.shape) == 2, "Array of shape (n_samples, n_timestamps) must be passed."
    n_samples, n_timestamps = a.shape

    a_binned = np.zeros(a.shape, dtype=np.int64)
    quantiles = np.linspace(0, 1, n_bins + 1)[1:-1]  # evenly spaced quantiles
    bins = np.quantile(a, q=quantiles, axis=1).T  # bins for each sample
    for i in prange(n_samples):
        a_binned[i, :] = np.searchsorted(
            bins[i, :], a[i, :], side="left"
        )  # squeeze coordinates

    mtm = np.zeros((n_samples, n_bins, n_bins))  # build Markov Transition Matrix
    for i in prange(n_samples):
        for j in prange(n_timestamps - 1):
            mtm[i, a_binned[i, j], a_binned[i, j + 1]] += 1

    mtm_sum = mtm.sum(axis=2)  # normalize rows sums in each sample to 1
    np.place(mtm_sum, mtm_sum == 0, 1)
    mtm /= mtm_sum[:, :, None]

    mtf = np.zeros(
        (n_samples, n_timestamps, n_timestamps)
    )  # build Markov Transition Field
    for i in prange(n_samples):
        for j in prange(n_timestamps):
            for k in prange(n_timestamps):
                mtf[i, j, k] = mtm[i, a_binned[i, j], a_binned[i, k]]

    return mtf


def _shrink_matrix(
    mat: np.array,
    height: int,
    width: int,
    strategy: Literal["max", "mean", "normal"] = "normal",
) -> np.array:
    """
    Shrinks matrix to be close to output_size x output_size.
    :param mat: 2d matrix (image channel).
    :param height: height of shrunk matrix.
    :param width: width of shrunk matrix.
    :param strategy: strategy to use while shrinking.
            * 'mean' - 2d convolution with uniform kernel.
            * 'normal' - 2d convolution with Gauss kernel.
            * 'max' - max pooling, resulting image is the closest possible to provided 'size'.
    """
    ih, iw = mat.shape
    oh, ow = height, width

    assert len(mat.shape) == 2
    assert oh < ih and ow < iw

    if strategy in ("mean", "normal"):  # convolution (uniform/normal)
        dh, dw = ih - oh + 1, iw - ow + 1
        if strategy == "mean":
            kernel = np.ones((dh, dw)) / (dh * dw)
        else:  # 'normal'
            m = max(dh, dw)
            kernel = _gaussian_kernel(m, sigma=5)
            kernel = _rec2square(kernel)

        shrunk_mat = convolve2d(mat, kernel, mode="valid")

    elif strategy == "max":  # max pooling
        dh, dw = int(np.ceil(ih / oh)), int(np.ceil(iw / ow))
        rh, rw = ih // dh, iw // dw

        if ih % dh == 0 and iw % dw == 0:  # filter divides matrix on equal blocks
            shrunk_mat = mat.reshape(rh, dh, rw, dw).max(axis=(1, 3))
        else:
            Q1 = mat[: rh * dh, : rw * dw].reshape(rh, dh, rw, dw).max(axis=(1, 3))
            Q2 = mat[rh * dh :, : rw * dw].reshape(-1, rw, dw).max(axis=2)
            Q3 = mat[: rh * dh, rw * dw :].reshape(rh, dh, -1).max(axis=1)
            Q4 = mat[rh * dh :, rw * dw :].max()
            shrunk_mat = np.vstack(np.c_[Q1, Q3], np.c_[Q2, Q4])

    else:
        raise ValueError(f"'shrink_strategy'={strategy} is not supported.")

    return shrunk_mat


def _gaussian_kernel(size, sigma) -> np.array:
    if size % 2 == 0:
        x = (np.arange(-size / 2, size / 2, 1) + 0.5).reshape(1, -1)
    else:
        x = np.arange(-size // 2 + 1, size // 2 + 1, 1).reshape(1, -1)

    xx = np.tile(x, (size, 1))
    yy = np.rot90(xx, 1)

    kernel = (1 / 2 * np.pi * np.square(sigma)) * np.exp(
        -(np.square(xx) + np.square(yy)) / (2 * np.square(sigma))
    )
    kernel = kernel / np.sum(kernel)
    return kernel


# =========================== GASF/GADF ===========================
def get_gaf(
    data: pd.DataFrame,
    x: str,
    y: str,
    t: str = None,
    field_type: Literal["difference", "sum"] = "difference",
    to_polar: Literal["regular", "cosine"] = "cosine",
    flatten: bool = False,
) -> np.array:
    """
    Calculates Gramian Angular Field for (x,y) coordinates.

    :param data: dataframe containing fixation coordinates.
    :param x: x-coordinate column name.
    :param y: y-coordinate column name.
    :param t: timestamps column name.
    :param field_type: which type of field to calculate. If "difference", then GADF is returned,
                       otherwise ("sum") GASF is returned.
    :param to_polar: conversion from cartesian to polar coordinates.
                    * 'regular': standard conversion calculating arctan(y/x).
                    * 'cosine': angle is calculated as cosine of series data, radius is taken as timestamps.
    :param flatten: bool, whether to flatten the array.

    :returns: tensor of shape (2, n_coords, n_coords), where n_coords is the length of input dataframe.
    """
    assert field_type in (
        "difference",
        "sum",
    ), f"'field_type'={field_type} is not supported."
    assert to_polar in ("regular", "cosine"), f"'to_polar'={to_polar} is not supported."
    x_coords, y_coords = data[[x]].values.ravel(), data[[y]].values.ravel()
    timestamps = np.arange(len(x_coords)) if t is None else data[[t]].values.ravel()
    gaf = _get_gaf(
        np.array([x_coords, y_coords]),
        timestamps,
        field_type=field_type,
        to_polar=to_polar,
    )
    return gaf.flatten() if flatten else gaf


def _get_gaf(
    a: np.array,
    t: np.array,
    field_type: Literal["difference", "sum"],
    to_polar: Literal["regular", "cosine"],
) -> np.array:
    """
    :param a: array of shape (n_samples, n_timestamps) being angular values.
    :param t: array of shape (n_timestamps,) being timestamps for all samples in 'a' or
                             (n_samples, n_timestamps) being timestamps of corresponding angular values in 'a',
                                                       i.e. a[i, :] are expected to be the angles at time t[i, :].
    """

    def _get_t(t_: np.array, i_: int):
        return t_[i_, :] if len(t_.shape) > 1 else t_

    assert len(a.shape) == 2, "Array of shape (n_samples, n_timestamps) must be passed."
    if len(t.shape) == 2:
        assert t.shape == a.shape, "'a' and 't' must be of same shape."
    else:
        assert len(t.shape) == 1 and len(t) == a.shape[1]

    n_samples, n_timestamps = a.shape
    if field_type == "sum":

        def f(phi1, phi2):
            return np.cos(phi1 + phi2)

    else:  # 'difference'

        def f(phi1, phi2):
            return np.sin(phi1 - phi2)

    gaf = np.zeros((n_samples, n_timestamps, n_timestamps))
    for i in prange(n_samples):
        if to_polar == "regular":
            rho, phi = _car2pol(
                a[i, :], _get_t(t, i)
            )  # _get_t used to avoid copies of 't' in 1d case
        else:
            rho, phi = _encode_car(a[i, :], _get_t(t, i))

        for j in prange(n_timestamps):
            for k in prange(n_timestamps):
                gaf[i, j, k] = f(phi[j], phi[k])

    return gaf


def _rescale(a: np.array) -> np.array:
    """
    Linearly map array to [-1, 1].
    """
    amm = _minmax(a)
    return amm * 2 - 1


def _minmax(a: np.array) -> np.array:
    min_, max_ = a.min(), a.max()
    if min_ == max_:
        return a if min_ == 0.0 else np.ones(a.shape, dtype=a.dtype)
    return (a - min_) / (max_ - min_)


def _car2pol(x: np.array, f_x: np.array) -> Tuple[np.array, np.array]:
    rho = np.sqrt(x**2 + f_x**2)
    phi = np.arctan2(f_x, x)
    return rho, phi


def _encode_car(x: np.array, t: np.array) -> Tuple[np.array, np.array]:
    rho = t
    phi = np.arccos(_rescale(x))
    return rho, phi


# =========================== HILBERT CURVE ===========================
@jit(forceobj=True, looplift=True)
def get_hilbert_curve_enc(
    data: pd.DataFrame, x: str, y: str, scale: bool = True, p: int = 4
) -> np.array:
    """
    Map scanpath to values on Hilbert curve and encode to single feature vector.

    :param data: dataframe containing fixation coordinates.
    :param x: x-coordinate column name.
    :param y: y-coordinate column name.
    :param scale: whether to scale the scanpath to [0, 1] before mapping to Hilbert curve. If false, then
    :param p: order of Hilbert curve, unit square is divided into (2^p)x(2^p) smaller squares.
              Higher value of p indicates better locality preservation.
    :return: scanpath encoding in 2^p-dimensional feature space using 1D Hilbert curve.
    """
    n = 2**p
    mapping = get_hilbert_curve(
        data=data, x=x, y=y, scale=scale, p=p
    )  # get Hilbert curve mapping
    mapping = np.unique(mapping)  # leave only unique values
    vec = np.zeros((n * n,))
    for i in prange(n * n):
        vec[i] = i in mapping  # activate mapped values
    return vec


@jit(forceobj=True, looplift=True)
def get_hilbert_curve(
    data: pd.DataFrame, x: str, y: str, scale: bool = True, p: int = 4
) -> np.array:
    """
    Map scanpath to points on 1D Hilbert curve.

    :param data: dataframe containing fixation coordinates.
    :param x: x-coordinate column name.
    :param y: y-coordinate column name.
    :param scale: whether to scale the scanpath to [0, 1] before mapping to Hilbert curve. If false, then
    :param p: order of Hilbert curve, unit square is divided into (2^p)x(2^p) smaller squares.
              Higher value of p indicates better locality preservation.
    :return: scanpath mapping to 1D Hilbert curve.
    """
    x, y = data[x].values, data[y].values
    n_fixations = len(x)

    if scale:
        x, y = _minmax(x), _minmax(
            y
        )  # map x, y to [0, 1]  TODO: better approach than minmax?
    else:
        assert (
            0 <= x <= 1
        ), "Either scale 'x' to be between 0 and 1 or add 'scale'=True."
        assert (
            0 <= y <= 1
        ), "Either scale 'y' to be between 0 and 1 or add 'scale'=True."
    x, y = x * (2**p), y * (2**p)  # map x, y to [0, 2^p]
    x, y = np.array(np.round(x), dtype=int), np.array(
        np.round(y), dtype=int
    )  # map [0, 2^p] to {0, 1, .., 2^p}

    h = np.zeros((n_fixations,))
    for i in prange(n_fixations):
        h[i] = xy2h(x[i], y[i], p=p)
    return h


def xy2h(x: int, y: int, p: int) -> int:
    """
    Mapping of 2D space to 1D using Hilbert curve.

    :param x: x-coordinate of a point, 0 <= x < 2^p.
    :param y: y-coordinate of a point, 0 <= y < 2^p.
    :param p: order of Hilbert curve, unit square is divided into (2^p)x(2^p) smaller squares.
              Higher value of p indicates better locality preservation.
    :return: corresponding point on 1D Hilbert curve.

    Algorithm: https://people.math.sc.edu/Burkardt/py_src/hilbert_curve/hilbert_curve.py.
    """
    n = 2**p

    s = n // 2
    d = 0

    while s > 0:
        rx = (x & s) > 0
        ry = (y & s) > 0
        d += s * s * ((3 * rx) ^ ry)  # adding length on Hilbert curve
        if ry == 0:  # rotation of quadrant
            if rx == 1:
                x = n - 1 - x
                y = n - 1 - y
            x, y = y, x
        s = s // 2
    return d


def hilbert_huang_transform(data: np.ndarray, max_imf: int = 1) -> np.ndarray:
    """
    Perform Hilbert-Huang transform on a given data sequence.

    :param data: input sequence of form (x, y) coordinates
    :param max_imf: maximum number of intrinsic mode functions to extract
    :return: np.ndarray: intrinsic mode functions from an input data vector
    """

    emd = EMD2D()
    decomposed = emd(data, max_imf=max_imf)
    return decomposed


# =========================== PERSISTENCE CURVE ===========================


def vietoris_rips_filtration(
    scanpath: np.array, max_dim: int = 2, max_radius: float = 1.0
):
    """
    Compute the Vietoris-Rips filtration for a point cloud.

    :param point_cloud: scanpath data as a numpy array of shape (n, 2)
    :param max_dim: maximum dimension for persistent homology
    :param max_radius: maximum radius for the filtration
    :return: persistence diagram and simplex tree
    """

    rips_complex = gd.RipsComplex(points=scanpath, max_edge_length=max_radius)
    simplex_tree = rips_complex.create_simplex_tree(max_dimension=max_dim)
    persistence = simplex_tree.persistence()

    return persistence, simplex_tree


def lower_star_filtration(time_series: np.ndarray, max_dim: int = 1):
    """
    Compute the Lower Star filtration for a time series.

    :param time_series: time series data
    :param max_dim: maximum dimension for persistent homology
    :param persistence: persistence diagram.
    """

    simplex_tree = gd.SimplexTree()

    for i, val in enumerate(time_series):
        simplex_tree.insert([i], filtration=val)

    simplex_tree.make_filtration_non_decreasing()
    persistence = simplex_tree.persistence()

    return persistence, simplex_tree


def persistence_curve(persistence_diagram: List[Tuple | np.ndarray], t: float):
    """
    Compute the persistence curve for a persistence diagram at time t.

    :param persistence_diagram: persistence diagram [(birth, death), ...]
    :param t: threshold time for persistence curve
    :return: sum of persistence intervals active at time t
    """
    total_persistence = sum(
        (death - birth) for birth, death in persistence_diagram if birth <= t <= death
    )
    return total_persistence


def persistence_entropy_curve(persistence_diagram: List[Tuple | np.ndarray], t: float):
    """
    Compute the persistence entropy curve for a persistence diagram at time t.

    :param persistence_diagram: persistence diagram [(birth, death), ...]
    :param t: threshold time for persistence entropy curve
    :return: entropy value at time t
    """

    intervals = [
        (death - birth) for birth, death in persistence_diagram if birth <= t <= death
    ]
    total_persistence = sum(intervals)

    if total_persistence == 0:
        return 0

    probabilities = [interval / total_persistence for interval in intervals]
    entropy = -sum(p * np.log(p) for p in probabilities)

    return entropy


def calculate_topological_features(
    scanpath: np.ndarray,
    time_series: np.ndarray,
    max_radius: float = 1.0,
    max_time: float = 1.0,
    max_dim: int = 2,
    time_steps: int = 100,
):
    """
    Calculate topological features (persistence curve and persistence entropy) for a scanpath.

    :param scanpath: scanpath data array of shape (n, 2)
    :param time_series: 1d array of time series data (e.g. x or y coordinates over time)
    :param max_radius: maximum radius for Vietoris-Rips filtration
    :param max_time: maximum threshold for the persistence curve
    :param max_dim: maximum dimension for persistent homology
    :param time_steps: number of time steps for evaluating the persistence curve and entropy
    :return: values of persistence curve at each time step, Values of persistence entropy at each time step
    """

    persistence_vr, _ = vietoris_rips_filtration(
        scanpath, max_dim=max_dim, max_radius=max_radius
    )
    persistence_ls, _ = lower_star_filtration(time_series, max_dim=max_dim)

    time_grid = np.linspace(0, max_time, time_steps)
    persistence_curve_vals = []
    persistence_entropy_vals = []

    for t in time_grid:
        pc_vr = persistence_curve([p[1] for p in persistence_vr], t)
        pe_vr = persistence_entropy_curve([p[1] for p in persistence_vr], t)

        pc_ls = persistence_curve([p[1] for p in persistence_ls], t)
        pe_ls = persistence_entropy_curve([p[1] for p in persistence_ls], t)

        persistence_curve_vals.append(pc_vr + pc_ls)
        persistence_entropy_vals.append(pe_vr + pe_ls)

    return np.array(persistence_curve_vals), np.array(persistence_entropy_vals)



import io
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from PIL import Image

from eyefeatures.utils import _select_regressions


def _built_figure(
    fig_dict: Dict, animation_duration: int = 500
):  # animation_duration in ms
    """
    Function for building a layout for plot
    """
    fig_dict["layout"]["width"] = 600
    fig_dict["layout"]["height"] = 600
    fig_dict["layout"]["updatemenus"] = [
        {
            "buttons": [
                {
                    "args": [
                        None,
                        {
                            "frame": {"duration": animation_duration, "redraw": False},
                            "fromcurrent": True,
                            "transition": {
                                "duration": 300,
                                "easing": "quadratic-in-out",
                            },
                        },
                    ],
                    "label": "Play",
                    "method": "animate",
                },
                {
                    "args": [
                        [None],
                        {
                            "frame": {"duration": 0, "redraw": False},
                            "mode": "immediate",
                            "transition": {"duration": 0},
                        },
                    ],
                    "label": "Pause",
                    "method": "animate",
                },
            ],
            "direction": "left",
            "pad": {"r": 10, "t": 87},
            "showactive": False,
            "type": "buttons",
            "x": 0.1,
            "xanchor": "right",
            "y": 0.0,
            "yanchor": "top",
        }
    ]


def tracker_animation(
    data_: pd.DataFrame,
    x: str,
    y: str,
    path_color: str = "green",
    path_width: float = 0.5,
    points_color: str = "black",
    points_width: float = 6,
    add_regression: bool = False,
    regression_color: str = "red",
    meta_data: List[str] = None,
    rule: Tuple[int, ...] = None,
    deviation: Union[int, Tuple[int, ...]] = None,
    aoi: str = None,
    aoi_c: Dict[str, str] = None,
    tracker_color: str = "red",
    animation_duration: int = 500,
    save_gif: str = None,
    frames_count: int = 1,
):
    """
    Function for tracker animation

    :param data_: DataFrame with fixations.
    :param x: x coordinate of fixation.
    :param y: y coordinate of fixation.
    :param path_color: color of saccades.
    :param path_width: width of saccades.
    :param points_color: color of points.
    :param points_width: width of points.
    :param add_regression: whether to add regressions.
    :param regression_color: color of regressions.
    :param meta_data: list of columns that will be used for meta data.
    :param rule: must be either 1) tuple of quadrants direction to classify
            regressions, 1st quadrant being upper-right square of plane and counting
            anti-clockwise or 2) tuple of angles in degrees (0 <= angle <= 360).
    :param deviation: if None, then `rule` is interpreted as quadrants. Otherwise,
            `rule` is interpreted as angles. If integer, then is a +-deviation for all angles.
            If tuple of integers, then must be of the same length as `rule`, each value being
            a corresponding deviation for each angle. Angle = 0 is positive x-axis direction,
            rotating anti-clockwise.
    :param aoi: AOI of fixations.
    :param aoi_c: colormap for AOI.
    :param tracker_color: color of tracker.
    :param animation_duration: duration of animation.
    :param save_gif: path to save animation.
    :param frames_count: TODO.
    """
    data = data_.reset_index(drop=True)
    X = data[x].values
    Y = data[y].values
    dX = data[x].diff()
    dY = data[y].diff()
    indexes = data.index

    fig_dict = {"data": [], "layout": {}, "frames": []}

    _built_figure(fig_dict, data.shape[0], animation_duration)

    sliders_dict = {
        "active": 0,
        "yanchor": "top",
        "xanchor": "left",
        "currentvalue": {
            "font": {"size": 20},
            "prefix": "Index:",
            "visible": True,
            "xanchor": "right",
        },
        "transition": {"duration": 300, "easing": "cubic-in-out"},
        "pad": {"b": 10, "t": 50},
        "len": 0.9,
        "x": 0.1,
        "y": 0.0,
        "steps": [],
    }

    edges = {
        "x": X,
        "y": Y,
        "mode": "lines",
        "line": dict(color=path_color, width=path_width),
        "name": "saccades",
    }

    if not (aoi is None) and aoi_c is None:
        aoi_c = dict()
        areas = data[aoi].unique()
        for area in areas:
            color = (
                np.random.randint(0, 255),
                np.random.randint(0, 255),
                np.random.randint(0, 255),
            )
            aoi_c[area] = f"rgb({color[0]}, {color[1]}, {color[2]})"

    if add_regression:
        mask = _select_regressions(dX, dY, rule, deviation)
        reg_ind_x = dX[mask].index
        reg_ind = []
        for i in reg_ind_x:
            if i != 0:
                reg_ind.append(i - 1)
                reg_ind.append(i)
        data["is_reg"] = [1 if z in reg_ind else 0 for z in data.index]
        edges = []
        first_reg = True
        first_sac = True
        for i in range(1, len(data)):
            if data.loc[i - 1, "is_reg"] == data.loc[i, "is_reg"] == 1:
                edges.append(
                    {
                        "x": [data.loc[i - 1, x], data.loc[i, x]],
                        "y": [data.loc[i - 1, y], data.loc[i, y]],
                        "mode": "lines",
                        "line": dict(color=regression_color, width=path_width),
                        "name": "regressions",
                        "showlegend": first_reg,
                    }
                )
                first_reg = False
            else:
                edges.append(
                    {
                        "x": [data.loc[i - 1, x], data.loc[i, x]],
                        "y": [data.loc[i - 1, y], data.loc[i, y]],
                        "mode": "lines",
                        "line": dict(color=path_color, width=path_width),
                        "name": "saccades",
                        "showlegend": first_sac,
                    }
                )
                first_sac = False
        fig_dict["data"].extend(edges)
    else:
        fig_dict["data"].append(edges)

    if not (aoi is None):
        areas = data[aoi].unique()
        nodes = []
        for area in areas:
            annotate = []
            data_area = data[data[aoi] == area]
            indexes_area = data_area.index
            for i in range(len(indexes_area)):
                if not (meta_data is None):
                    row = data_area.loc[
                        indexes_area[i], data_area.columns.intersection(meta_data)
                    ].values
                    comments = []
                    for j in range(len(meta_data)):
                        comments.append(f"{meta_data[j]}: {row[j]}")
                    annotate.append("<br>".join(comments))
            nodes.append(
                {
                    "x": data_area[x].values,
                    "y": data_area[y].values,
                    "mode": "markers",
                    "marker": dict(color=aoi_c[area], size=points_width),
                    "name": area,
                    "text": annotate,
                }
            )
        fig_dict["data"].extend(nodes)
    else:
        annotate = []
        for i in range(len(indexes)):
            if not (meta_data is None):
                row = data.loc[indexes[i], data.columns.intersection(meta_data)].values
                comments = []
                for j in range(len(meta_data)):
                    comments.append(f"{meta_data[j]}: {row[j]}")
                annotate.append("<br>".join(comments))
        nodes = {
            "x": X,
            "y": Y,
            "mode": "markers",
            "marker": dict(color=points_color, size=points_width),
            "name": "fixations",
            "text": annotate,
        }
        fig_dict["data"].append(nodes)

    fig_dict["data"].append(
        {
            "x": [X[0]],
            "y": [Y[0]],
            "mode": "markers",
            "marker": dict(color=tracker_color),
            "name": "tracker",
        }
    )
    gif_list = []
    for i in range(len(indexes)):
        frame = {"data": [], "name": str(i)}
        if add_regression:
            frame["data"].extend(edges)
        else:
            frame["data"].append(edges)
        if not (aoi is None):
            frame["data"].extend(nodes)
        else:
            frame["data"].append(nodes)
        frame["data"].append(
            {
                "x": [X[i]],
                "y": [Y[i]],
                "mode": "markers",
                "marker": dict(color=tracker_color),
                "name": "tracker",
            }
        )
        fig_dict["frames"].append(frame)
        if not (save_gif is None):
            img = Image.open(io.BytesIO(go.Figure(frame).to_image(format="png")))
            for _ in range(frames_count):
                gif_list.append(img)
        slider_step = {
            "args": [
                [str(i)],
                {
                    "frame": {"duration": 300, "redraw": False},
                    "mode": "immediate",
                    "transition": {"duration": 300},
                },
            ],
            "label": str(i),
            "method": "animate",
        }
        sliders_dict["steps"].append(slider_step)

    fig_dict["layout"]["sliders"] = [sliders_dict]
    fig = go.Figure(fig_dict)
    fig.show()
    if not (save_gif is None):
        gif_list[0].save(
            save_gif,
            save_all=True,
            append_images=gif_list[1:],
            durarion=1000,
            loop=0,
            fps=1,
        )


def scanpath_animation(
    data_: pd.DataFrame,
    x: str,
    y: str,
    path_color: str = "green",
    path_width: float = 0.5,
    points_color: str = "black",
    points_width: float = 6,
    add_regression: bool = False,
    regression_color: str = "red",
    rule: Tuple[int, ...] = None,
    deviation: Union[int, Tuple[int, ...]] = None,
    animation_duration: int = 500,
    save_gif: str = None,
    frames_count: int = 1,
):
    """
    Function for tracker animation

    :param data_: DataFrame with fixations.
    :param x: x coordinate of fixation.
    :param y: y coordinate of fixation.
    :param path_color: color of saccades.
    :param path_width: width of saccades.
    :param points_color: color of points.
    :param points_width: width of points.
    :param add_regression: whether to add regressions.
    :param regression_color: color of regressions.
    :param rule: must be either 1) tuple of quadrants direction to classify
            regressions, 1st quadrant being upper-right square of plane and counting
            anti-clockwise or 2) tuple of angles in degrees (0 <= angle <= 360).
    :param deviation: if None, then `rule` is interpreted as quadrants. Otherwise,
            `rule` is interpreted as angles. If integer, then is a +-deviation for all angles.
            If tuple of integers, then must be of the same length as `rule`, each value being
            a corresponding deviation for each angle. Angle = 0 is positive x-axis direction,
            rotating anti-clockwise.
    :param animation_duration: duration of animation.
    :param save_gif: path to save animation.
    :param frames_count: TODO.
    """
    data = data_.reset_index(drop=True)
    X = data[x].values
    Y = data[y].values
    x_min, x_max, y_min, y_max = X.min(), X.max(), Y.min(), Y.max()
    indexes = data.index

    fig_dict = {"data": [], "layout": {}, "frames": []}

    _built_figure(fig_dict, data.shape[0], animation_duration)

    sliders_dict = {
        "active": 0,
        "yanchor": "top",
        "xanchor": "left",
        "currentvalue": {
            "font": {"size": 20},
            "prefix": "Index:",
            "visible": True,
            "xanchor": "right",
        },
        "transition": {"duration": 300, "easing": "cubic-in-out"},
        "pad": {"b": 10, "t": 50},
        "len": 0.9,
        "x": 0.1,
        "y": 0.0,
        "steps": [],
    }

    fig_dict["layout"]["xaxis"] = dict(
        range=[x_min - 0.1, x_max + 0.1], automargin=False
    )
    fig_dict["layout"]["yaxis"] = dict(
        range=[y_min - 0.1, y_max + 0.1], automargin=False
    )

    fig_dict["data"].extend(
        [
            {
                "x": [],
                "y": [],
                "mode": "lines",
                "line": dict(color=path_color, width=path_width),
            }
            for _ in range(len(indexes))
        ]
    )

    graph = []

    if add_regression:
        dX = data[x].diff()
        dY = data[y].diff()
        mask = _select_regressions(dX, dY, rule, deviation)
        reg_ind_x = dX[mask].index
        reg_ind = []
        for i in reg_ind_x:
            if i != 0:
                reg_ind.append(i - 1)
                reg_ind.append(i)
        data["is_reg"] = [1 if z in reg_ind else 0 for z in data.index]

    gif_list = []
    for i in range(1, len(indexes)):
        frame = {"data": [], "name": str(i), "layout": {"xaxis": {}, "yaxis": {}}}
        color = path_color
        name = "saccedes"
        if add_regression and (data.loc[i - 1, "is_reg"] == data.loc[i, "is_reg"] == 1):
            color = regression_color
            name = "regression"
        new_edge = {
            "x": [data.loc[i - 1, x], data.loc[i, x]],
            "y": [data.loc[i - 1, y], data.loc[i, y]],
            "mode": "lines+markers",
            "line": dict(color=color, width=path_width),
            "marker": dict(color=points_color, size=points_width),
            "name": name,
            "showlegend": False,
        }
        frame["data"].extend(graph)
        frame["data"].append(new_edge)
        frame["layout"]["xaxis"] = dict(
            range=[x_min - 0.1, x_max + 0.1], automargin=False
        )
        frame["layout"]["yaxis"] = dict(
            range=[y_min - 0.1, y_max + 0.1], automargin=False
        )
        fig_dict["frames"].append(frame)
        if not (save_gif is None):
            img = Image.open(io.BytesIO(go.Figure(frame).to_image(format="png")))
            for _ in range(frames_count):
                gif_list.append(img)
        graph.append(new_edge)
        slider_step = {
            "args": [
                [str(i)],
                {
                    "frame": {"duration": 300, "redraw": False},
                    "mode": "immediate",
                    "transition": {"duration": 300},
                },
            ],
            "label": str(i),
            "method": "animate",
        }
        sliders_dict["steps"].append(slider_step)

    fig_dict["layout"]["sliders"] = [sliders_dict]
    fig = go.Figure(fig_dict)
    fig.show()
    if not (save_gif is None):
        gif_list[0].save(
            save_gif,
            save_all=True,
            append_images=gif_list[1:],
            durarion=1000,
            fps=1,
            loop=0,
        )



from typing import List, Union

import numpy as np
from numpy.typing import NDArray


def _get_distance(
    v: Union[float, NDArray], u: Union[float, NDArray], distance: str
) -> Union[float, NDArray]:
    """
    Method computes vectors norm given distance.

    :param v: number/vector/matrix.
    :param u: number/vector/matrix.
    :param distance: str, distance to compute.
    :return: computed distance, number/vector.

    Note
    -------
        If u and v are numbers, then all distances are equivalent to |v - u|.
        If u and v are vectors of size d, then corresponding metric in R^d is returned.
        If u and v are matrices, then must be of size n x d, rows are treated as vectors.

        All distances are norms in Euclidean space.
    """
    is_matrix = isinstance(v, np.ndarray) and len(v.shape) > 1
    if distance == "euc":
        return _euc_distance(v, u, is_matrix)
    if distance == "manhattan":
        return _manhattan_distance(v, u, is_matrix)
    if distance == "chebyshev":
        return _chebyshev_distance(v, u, is_matrix)

    raise NotImplementedError(f"Provided distance '{distance}' is not supported.")


# Warning: All distances are norms in Euclidean space. If this changes,
# modify the docstring of _get_distance accordingly.


def _euc_distance(
    v: Union[float, NDArray], u: Union[float, NDArray], is_matrix
) -> Union[float, NDArray]:
    if not is_matrix:
        return np.sqrt(np.sum(np.square(v - u)))
    else:
        return np.sqrt(np.sum(np.square(v - u), axis=1))


def _manhattan_distance(
    v: Union[float, NDArray], u: Union[float, NDArray], is_matrix
) -> Union[float, NDArray]:
    if not is_matrix:
        return np.sum(np.abs(v - u))
    else:
        return np.sum(np.abs(v - u), axis=1)


def _chebyshev_distance(
    v: Union[float, NDArray], u: Union[float, NDArray], is_matrix
) -> Union[float, NDArray]:
    if not is_matrix:
        return np.max(np.abs(v - u))
    else:
        return np.max(np.abs(v - u), axis=1)


# ====== Bounding Boxes =========
def _is_valid_circle(points: NDArray, center: NDArray, radius: float):
    for p in points:
        if np.linalg.norm(p - center) > radius:
            return False
    return True


def _build_min_circle(points: NDArray):
    assert points.shape[0] <= 3
    if points.shape[0] == 0:
        return np.array([0, 0, 0])
    if points.shape[0] == 1:
        return np.array([points[0][0], points[0][1], 0])
    if points.shape[0] == 2:
        x = (points[0][0] + points[1][0]) / 2
        y = (points[0][1] + points[1][1]) / 2
        r = np.linalg.norm(points[0] - points[1]) / 2
        return np.array([x, y, r])

    for i in range(points.shape[0]):
        for j in range(i + 1, points.shape[0]):
            x = (points[i][0] + points[j][0]) / 2
            y = (points[i][1] + points[j][1]) / 2
            r = np.linalg.norm(points[i] - points[j]) / 2
            if _is_valid_circle(points, np.array([x, y]), r):
                return np.array([x, y, r])
    # define circle by 3 points
    bx = points[1][0] - points[0][0]
    by = points[1][1] - points[0][1]
    cx = points[2][0] - points[0][0]
    cy = points[2][1] - points[0][1]
    B = bx * bx + by * by
    C = cx * cx + cy * cy
    D = bx * cy - by * cx
    assert D != 0.0, f"Division by zero"
    x = (cy * B - by * C) / (2 * D) + points[0][0]
    y = (bx * C - cx * B) / (2 * D) + points[0][1]
    r = np.linalg.norm(np.array([x, y]) - points[0])
    return np.array([x, y, r])


def _welzl_algorithm(points: NDArray, border: List[int], N: int):
    if N == 0 or len(border) == 3:
        return _build_min_circle(np.array(border))

    d = _welzl_algorithm(points, border.copy(), N - 1)

    if np.linalg.norm(d[:2] - points[N - 1]) <= d[2]:
        return d
    border.append(points[N - 1])

    return _welzl_algorithm(points, border.copy(), N - 1)


def _get_MEC(points: NDArray):
    points = points.copy()
    return _welzl_algorithm(points, [], len(points))



from typing import List, Literal, Tuple, Union

import numpy as np
import pandas as pd
from scipy.signal import firwin, iirfilter, lfilter, savgol_filter

from eyefeatures.preprocessing.base import BaseSmoothingPreprocessor


# ======== SMOOTHING PREPROCESSORS ========
class SavGolFilter(
    BaseSmoothingPreprocessor
):  # TODO 2D SG filter: https://github.com/espdev/sgolay2
    """
    Savitzkiy-Golay filter. 'x' and 'y' directions are filtered independently, time is ignored.
    Parameters are passed to `scipy.signal.savgol_filter`.
    \n
    Default values are taken from https://arxiv.org/pdf/2303.02134.
    """

    def __init__(
        self,
        x: str,
        y: str,
        t: str = None,
        pk: List[str] = None,
        window_length: int = 11,
        polyorder: int = 2,
        **savgol_kw
    ):
        super().__init__(x=x, y=y, t=t, pk=pk)
        self.wl = window_length
        self.po = polyorder
        self.savgol_kw = savgol_kw

    def _check_params(self):
        m = "Savitzkiy-Golay filter"
        assert self.x is not None, self._err_no_field(m, "x")
        assert self.y is not None, self._err_no_field(m, "y")

    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        if self.pk:
            X = X.drop(self.pk, axis=1)

        X_filt = X.copy()
        if len(X_filt) >= self.wl:
            X_filt[self.x] = savgol_filter(
                x=X[self.x].values,
                window_length=self.wl,
                polyorder=self.po,
                **self.savgol_kw
            )
            X_filt[self.y] = savgol_filter(
                x=X[self.y].values,
                window_length=self.wl,
                polyorder=self.po,
                **self.savgol_kw
            )
        return X_filt


class FIRFilter(BaseSmoothingPreprocessor):  # TODO 2D version?
    """
    FIR filter. Convolution with RIR kernel along 'x' and 'y'. `kwargs` are passed to `scipy.signal.firwin`
    to determine the kernel.

    :param mode: parameter of `scipy.signal.fftconvolve`.

    Note
    -------
    Default values are taken from https://arxiv.org/pdf/2303.02134.
    """

    def __init__(
        self,
        x: str,
        y: str,
        t: str = None,
        pk: List[str] = None,
        numtaps: int = 81,
        fs: int = 250,
        cutoff: Union[float, Tuple[float, ...]] = 100,
        pass_zero: Literal[
            False, True, "bandpass", "lowpass", "highpass", "bandstop"
        ] = False,
        mode: Literal["valid", "full", "same"] = "valid",
        **fir_kw
    ):
        super().__init__(x=x, y=y, t=t, pk=pk)
        self.mode = mode
        self.numtaps = numtaps
        self.cutoff = cutoff
        self.pass_zero = pass_zero
        self.fs = fs
        self.fir_kw = fir_kw

    def _check_params(self):
        m = "FIR filter"
        assert self.x is not None, self._err_no_field(m, "x")
        assert self.y is not None, self._err_no_field(m, "y")

    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        if self.pk:
            X = X.drop(self.pk, axis=1)

        X_filt = X.copy()
        if len(X) > self.numtaps:
            kernel = firwin(
                numtaps=self.numtaps,
                cutoff=self.cutoff,
                pass_zero=self.pass_zero,
                fs=self.fs,
                **self.fir_kw
            )
            x_filt = np.convolve(X[self.x].values.ravel(), kernel, mode=self.mode)
            y_filt = np.convolve(X[self.y].values.ravel(), kernel, mode=self.mode)

            # slice part of array based on convolution mode
            k, h = len(kernel) - 1, (len(kernel) - 1) // 2
            if self.mode == "valid":
                X_filt = X[h + (k % 2) : -h]
            elif self.mode == "same":
                pass
            else:  # "full"
                x_filt = x_filt[h + k % 2 : -h]
                y_filt = y_filt[h + k % 2 : -h]

            X_filt.loc[:, self.x] = x_filt
            X_filt.loc[:, self.y] = y_filt

        return X_filt


class IIRFilter(BaseSmoothingPreprocessor):  # TODO 2D version?
    """
    IIR filter. Convolution with IIR kernel along 'x' and 'y'. `kwargs` are passed to `scipy.signal.iirfilter`
    to determine the kernel.

    :param mode: parameter of `scipy.signal.fftconvolve`.

    Note
    -------
    Default values are taken from https://arxiv.org/pdf/2303.02134.
    """

    def __init__(
        self,
        x: str,
        y: str,
        t: str = None,
        pk: List[str] = None,
        N: int = 7,
        Wn: Union[int, Tuple[int, int]] = 0.5,
        **iir_kw
    ):
        super().__init__(x=x, y=y, t=t, pk=pk)
        self.N = N
        self.Wn = Wn
        self.iir_kw = iir_kw

    def _check_params(self):
        m = "IIR filter"
        assert self.x is not None, self._err_no_field(m, "x")
        assert self.y is not None, self._err_no_field(m, "y")
        assert (
            self.iir_kw.get("output", "ba") == "ba"
        ), "Only 'output'='ba' is supported."

    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        if self.pk:
            X = X.drop(self.pk, axis=1)

        if self.iir_kw.get("ftype") is None:
            self.iir_kw["ftype"] = "butter"

        if self.iir_kw.get("btype") is None:
            self.iir_kw["btype"] = "highpass"

        b, a = iirfilter(N=self.N, Wn=self.Wn, **self.iir_kw, output="ba")
        x_filt = lfilter(b, a, X[self.x].values)[len(b) - 1 :]  # "valid" in np.convolve
        y_filt = lfilter(b, a, X[self.y].values)[len(b) - 1 :]  # "valid" in np.convolve

        k, h = len(a) - 1, (len(a) - 1) // 2
        X_filt = X[h + (k % 2) : -h]

        X_filt.loc[:, self.x] = x_filt
        X_filt.loc[:, self.y] = y_filt
        return X_filt


class WienerFilter(BaseSmoothingPreprocessor):
    """
    Wiener filter. Applied independently along 'x' and 'y' axis.

    :param K: estimate of ratio of noise-to-signal spectra. If 'auto', such value of K
              on grid [-1e-3, 1e-3] with 2e3 values is chosen to minimize PSNR.
    :param sigma: std of Gaussian filter.
    :param size: length of Gaussian filter.
    """

    def __init__(
        self,
        x: str,
        y: str,
        t: str = None,
        pk: List[str] = None,
        K: Union[float, Literal["auto"]] = 4.3e-5,
        sigma: float = 0.2,
        size: int = 11,
    ):
        super().__init__(x=x, y=y, t=t, pk=pk)
        self.K = K
        self.sigma = sigma
        self.size = size

    def _check_params(self):
        assert self.K == "auto" or isinstance(self.K, float)

    @staticmethod
    def _gaussian_kernel(size, sigma):
        if size % 2 == 0:
            x = (np.arange(-size / 2, size / 2, 1) + 0.5).ravel()
        else:
            x = np.arange(-size // 2 + 1, size // 2 + 1, 1).ravel()

        # get 1D kernel using Gaussian PDF
        kernel = (
            1
            / (2 * np.pi * np.square(sigma))
            * np.exp(-np.square(x) / (2 * np.square(sigma)))
        )
        # normalize kernel
        kernel = kernel / np.sum(kernel)
        return kernel

    @staticmethod
    def _compute_psnr(seq1, seq2):
        """
        Peak-Signal-to-Noise Ratio. Must be applied to normalized coordinates with maximum value of 1.
        """
        mse = np.sum(np.square(seq1 - seq2)) / len(seq1)
        psnr = 20 * np.log10(2 / np.sqrt(mse))  # 2 = length of [-1, 1]
        return psnr

    @staticmethod
    def _kernel_fft(kernel: np.array, length: int):
        diff = length - len(kernel.ravel())
        l = diff // 2
        r = diff - l

        # pad the kernel
        kernel_padded = np.pad(kernel, pad_width=(l, r))
        # IFFT for center-based kernel
        kernel_adjusted = np.fft.ifftshift(kernel_padded)
        # FFT to adjust kernel to the correct position
        kernel_transformed = np.fft.fft(kernel_adjusted)
        return kernel_transformed

    def _wiener_filter(self, seq: np.array, kernel: np.array, K):
        H = self._kernel_fft(kernel, seq.shape[-1])
        G = np.fft.fft(seq)
        H_conj = np.conjugate(H)

        F_est = (H_conj * G) / (H * H_conj + K)
        f_est = np.fft.ifft(F_est)

        return np.abs(f_est)  # take modulus of complex numbers

    def _preprocess(self, X: pd.DataFrame) -> pd.DataFrame:
        if self.pk:
            X = X.drop(self.pk, axis=1)

        # check if input has enough length
        if len(X) <= self.size:
            return X

        # define sequences for Wiener filter
        x_path, y_path = X[self.x].values, X[self.y].values
        # get 1D Gaussian kernel
        kernel = self._gaussian_kernel(self.size, self.sigma)

        # find optimal value for K
        if self.K == "auto":
            best_psnr = -np.inf
            best_K = 0

            for K in list(np.linspace(-1e-3, 1e-3, 2001)):
                x_filt = self._wiener_filter(x_path, kernel=kernel, K=K)
                y_filt = self._wiener_filter(y_path, kernel=kernel, K=K)

                psnr = self._compute_psnr(x_filt, y_filt)
                if psnr > best_psnr:
                    best_psnr = psnr
                    best_K = K

            self.K = best_K

        x_filt = self._wiener_filter(x_path, kernel=kernel, K=self.K)
        y_filt = self._wiener_filter(y_path, kernel=kernel, K=self.K)

        X_filt = X.copy()
        X_filt.loc[:, self.x] = x_filt
        X_filt.loc[:, self.y] = y_filt
        return X_filt
